#! /usr/bin/env python
# -*- coding: utf-8 -*-
# This serves as interface to run a problem from CUTEst 

import sys
import argparse
import numpy as np
import pycutestmgr
import pysleqp
import json

def print_problem_info(SIF):
    with open('problems.json', 'r') as f:
        problems = json.load(f)

    print
    print "Comments in Problem File"
    print problems[SIF]['comment']
    if 'obj' in problems[SIF].keys():
        print
        for o in problems[SIF]['obj']:
            print "Objective Value stated in Problem File: %e" % o
    print

def runproblem(SIF, maxiter=100, verbosity=0, printcomment=False, log=False, lpsolver='gurobi', parametric=False, stdout=None):
    test = pysleqp.SLEQP()
    if not pycutestmgr.isCached(SIF):
        pycutestmgr.prepareProblem(SIF)
    prob = pycutestmgr.importProblem(SIF)
    info = prob.getinfo()

    if info['m'] == 0:
        info['cl'] = np.array([])
        info['cu'] = np.array([])
        
    def firstorder(x,userdata):
        (obj, grad) = prob.obj(x, True)
        (cons, jac) = prob.scons(x)
        return (obj, grad, cons, jac)

    test.set_problem(x = info['x'], xl = info['bl'], xu = info['bu'])
    test.set_problem(cl = info['cl'], cu = info['cu'])
    test.set_problem(firstorder = firstorder)
    test.set_problem(verbosity=verbosity)
    test.set_problem(maxiter=maxiter)
    test.set_problem(n = info['n'], m = info['m'])
    test.set_problem(log=log)

    if info['m'] == 0:
        hess_lgn_prod = lambda x,s,l,d,userdata: s*prob.hprod(d,x)
        hess_lgn_prod_re = lambda x,s,l,d,userdata: s*prob.hprod(d,x)
    else:
        def hess_lgn_prod(x, s, l, d, userdata):
            if s == 1.:
                return prob.hprod(d,x,l)
            else:
                return s*prob.hprod(d,x,1./s*l)
        def hess_lgn_prod_re(x, s, l, d, userdata):
            return s*prob.hprod(d)
        
    test.set_problem(hess_lgn_prod = hess_lgn_prod)
    test.set_problem(hess_lgn_prod_re = hess_lgn_prod_re)
    test.set_problem(lpsolver=lpsolver)
    test.set_problem(parametric=parametric)
    if stdout != None:
        test.set_problem(stdout=stdout)

    test.optimize()
    if printcomment:
        print_problem_info(SIF)
    return [test, None]
    
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Execute Problem from CUTEst')
    parser.add_argument('SIF', type=str, help='Name of SIF File for Problem')
    parser.add_argument('--maxiter', type=int, help='Maximum Number of Iterations', default=100)
    parser.add_argument('--verbosity', type=int, help='Verbosity', default=1)
    parser.add_argument('--printcomment', action="store_true", help='Print Comments in SIF File', default=False)
    parser.add_argument('--lpsolver', type=str, help='LP Solver', default='gurobi')
    parser.add_argument('--parametric', action="store_true", help='Parametric Trust Region Choice', default=False)
    parser.add_argument('--log', help='Log to file', default=False)
    args = parser.parse_args()
    runproblem(args.SIF, maxiter=args.maxiter, verbosity=args.verbosity, printcomment=args.printcomment, log=args.log, lpsolver=args.lpsolver, parametric=args.parametric)
