# This software is licensed according to the pySLEQP academic license
# see LICENSE

import numpy as np
import scipy, scipy.sparse
import pysleqp

solver = pysleqp.SLEQP()

# number of variables
n = 4
# number of constraints
m = 2

# callback function that computes objective and constraints and derivatives
def firstorder(x, userdata):
    # objective
    obj = x[0]*x[3]*np.sum(x[:2]) + x[2]

    # gradient of objective
    grad = np.empty([n])
    grad[0] = x[0]*x[3] + x[3]*np.sum(x[:2])
    grad[1] = x[0]*x[3]
    grad[2] = x[0]*x[3] + 1
    grad[3] = x[0]*np.sum(x[:2])

    # constraints
    cons = np.empty([m])
    cons[0] = np.prod(x)
    cons[1] = np.dot(x, x)

    # jacobian of constraints, first dense to construct
    jac = np.empty([m, n])
    jac[0, 0] = x[1]*x[2]*x[3]
    jac[0, 1] = x[0]*x[2]*x[3]
    jac[0, 2] = x[0]*x[1]*x[3]
    jac[0, 3] = x[0]*x[1]*x[2]
    jac[1, :] = 2*x

    # convert to sparse
    spjac = scipy.sparse.csc_matrix(jac)

    return obj, grad, cons, spjac

# callback function that computes hessian vector product of lagrange function
# s * D^2 f(x)*d + sum_i l[i] * D^2 c_i(x)*d
def hessprod(x, s, l, d, userdata):
    hp = np.empty([n])
    hp[0] = ( s*( 2*x[3]*d[0] + x[3]*d[1] + x[3]*d[2] + (2*x[0]+x[1]+x[2])*d[3] ) + 
              l[0]*( x[2]*x[3]*d[1] + x[1]*x[3]*d[2] + x[1]*x[2]*d[3] ) +
              l[1]*d[0] )
    hp[1] = ( s*( x[3]*d[0] + x[0]*d[1] ) + 
              l[0]*( x[2]*x[3]*d[0] + x[0]*x[3]*d[2] + x[0]*x[2]*d[3] ) )
    hp[2] = ( s*( x[3]*d[0] + x[0]*d[1] ) +
              l[0]*( x[1]*x[3]*d[0] + x[0]*x[3]*d[1] + x[0]*x[1]*d[3] ) +
              l[1]*d[2] )
    hp[3] = ( s*( (2*x[0]+x[1]+x[2])*d[0] + x[0]*d[1] + x[0]*d[2] ) +
              l[0]*( x[1]*x[2]*d[0] + x[0]*x[2]*d[1] + x[0]*x[1]*d[2] ) + 
              2*l[1]*d[3] )
    return hp
    
solver.set_problem(
        n  = n, m = m,
        x  = np.array([1.0, 5.0, 5.0, 1.0]), # start point
        xl = np.array([1.0, 1.0, 1.0, 1.0]), # lower bounds for variables
        xu = np.array([5.0, 5.0, 5.0, 5.0]), # upper bounds for variables
        cl = np.array([25.0, 40.0]),         # lower bounds for constraints
        cu = np.array([1e20, 40.0]),         # upper bounds for constraints
        firstorder = firstorder,             # callback for firstorder function evaluations
        hess_lgn_prod = hessprod,            # callback for hessian-vector-product
        hess_lgn_prod_re = hessprod,         # callback for hessian-vector-product reevalution with some x, s, l but possibly different d
        verbosity = 1                        # level of screen outputs. default = 0 --> no screen output at all
        )

solver.optimize()

print
print "Solution point"
print "x:                             ", solver.x
print "Multipliers:                   ", solver.multiplier_scale
print "Multipliers for simple bounds: ", solver.multiplier_simple_scale
