! This software is licensed according to the pySLEQP academic license
! see LICENSE

module ma57_wrapper

use iso_c_binding, only: c_double, c_long, c_int
use HSL_MA57_double
use HSL_ZD11_double

implicit none

contains

subroutine c_ma57(n, m, ne_A, val_A, row_A, col_A, vector, &
    initialization, finalization, info, precond) bind(c)
integer(c_int), intent(in) :: n, m, ne_A, initialization, finalization
integer(c_int), intent(out) :: info
real(c_double), dimension(ne_A), intent(in) :: val_A
integer(c_long), dimension(ne_A), intent(in) :: row_A, col_A
real(c_double), dimension(n+m), intent(inout) :: vector
real(c_double), dimension(n+m) :: rhs
real(c_double), intent(in) :: precond
type(zd11_type), save :: K
type(ma57_control), save :: control
type(ma57_ainfo), save :: ainfo
type(ma57_finfo), save :: finfo
type(ma57_sinfo), save :: sinfo
type(ma57_factors), save :: factors
integer :: i, iter

if(initialization /= 0) then
    ! input matrix A in COO format
    ! construct symmetric matrix K = [ 1, A^T; A, 0 ] with lower 
    ! half stored in COO format
    K%n = n+m
    K%ne = ne_A + n
    if ( allocated( K%val ) ) deallocate( K%val, K%row, K%col )
    allocate(K%val(ne_A+n), K%row(ne_A+n), &
        K%col(ne_A+n))
    ! copy matrix A to lower block, indices coming from python
    ! have to be shifted by 1
    do i = 1, ne_A
        K%val(i) = val_A(i)
        K%row(i) = row_A(i) + 1 + n
        K%col(i) = col_A(i) + 1
    end do
    ! create entries for scaled preconditioner unit matrix on upper block
    do i = 1, n
        K%val(ne_A+i) = precond
        K%row(ne_A+i) = i
        K%col(ne_A+i) = i
    end do
    ! initialize structures
    call MA57_initialize(factors, control)
    ! analyze matrix structure and pivot pattern
    call MA57_analyse(K, factors, control, ainfo)
    ! obtain factorization
    call MA57_factorize(K, factors, control, finfo)
    info = finfo%flag
    return
end if

if(finalization /= 0) then
    if ( allocated( K%val ) ) deallocate( K%val, K%row, K%col )
    call MA57_finalize(factors, control, info)
    return
end if

rhs = vector
iter = 1
call MA57_solve(K, factors, vector, control, sinfo, rhs, iter)
info = sinfo%flag
return

end subroutine c_ma57


end module ma57_wrapper
