# This software is licensed according to the pySLEQP academic license
# see LICENSE

import numpy as np
import time
cimport numpy as np
from libc.stdlib cimport malloc, free

cdef extern:
    void c_gltr(int *n, double *radius, double *f, double *X, double *R, double *VECTOR, int *initialization, int *num_iter, int *info, int *negative_curvature, double *rayleigh, double *multiplier)

    
def eqpsolve(int n, double tr_eqp, double[:] grad_eqp, h_prod, proj, walltime):
    
    # calling c_gltr
    # n:        number of unknowns
    # radius:   must be set on initial to radius of quadratic constraint
    # f:        holds estimate of optimal function value. must not be set
    #           on initial entry
    # X:        array of dimension n that holds estimate of the solution
    #           of the linear system. On initial entry, X need not to be
    #           set. On exit, X contains the current best estimate of the
    #           solution
    # R:        array of dimension n that holds Hx+c, on initial entry
    #           must contain c
    # VECTOR:   array of dimension n used to pass information

    cdef double radius, f, rayleigh
    X = np.empty([n])
    VECTOR = np.empty([n])
    R = np.empty([n])
    cdef double[:] X_view = X
    cdef double[:] VECTOR_view = VECTOR
    cdef double[:] R_view = R
    
    cdef double zerotime = time.clock()

    # initialize GLTR
    cdef int initialization, num_iter, info, negative_curvature
    initialization = 1
    radius = tr_eqp
    R_view[:] = grad_eqp

    # for sanity assign default values
    num_iter = 0
    info = 0
    negative_curvature = 0
    f = 0.0
    multiplier = 0.0
    c_gltr(&n, &radius, &f, &X_view[0], &R_view[0], &VECTOR_view[0], &initialization, &num_iter, &info, &negative_curvature, &rayleigh, &multiplier)
    initialization = 0

    # loop c_gltr until solution is reached
    while True:
        if time.clock() - zerotime > walltime:
            raise RuntimeError('GLTR Walltime exceeded, number of iterations')
        c_gltr(&n, &radius, &f, &X_view[0], &R_view[0], &VECTOR_view[0], &initialization, &num_iter, &info, &negative_curvature, &rayleigh, &multiplier)
        if info == 2 or info == 6:
            # projection, VECTOR <-- M^-1 * VECTOR
            proj(VECTOR)
            if not np.isfinite(VECTOR).all():
                return X, rayleigh, -12, num_iter
        elif info == 3 or info == 7:
            # matrix vector product, VECTOR <-- H_EQP * VECTOR
            if h_prod(VECTOR) < 0:
                return X, rayleigh, -13, num_iter
        elif info == 5:
            # restart with initial gradient, reset R to grad_eqp
            R_view[:] = grad_eqp
        else:
            return X, rayleigh, info, num_iter
