# This software is licensed according to the pySLEQP academic license
# see LICENSE

# -*- coding: utf-8 -*-

cimport cgurobi
import numpy as np
cimport numpy as np
import scipy, scipy.sparse

status_dict = {}
status_dict[ 1] = 'LOADED'
status_dict[ 2] = 'OPTIMAL'
status_dict[ 3] = 'INFEASIBLE'
status_dict[ 4] = 'INF_OR_UNBD'
status_dict[ 5] = 'UNBOUNDED'
status_dict[ 6] = 'CUTOFF'
status_dict[ 7] = 'ITERATION_LIMIT'
status_dict[ 8] = 'NODE_LIMIT'
status_dict[ 9] = 'TIME_LIMIT'
status_dict[10] = 'SOLUTION_LIMIT'
status_dict[11] = 'INTERRUPTED'
status_dict[12] = 'NUMERIC'
status_dict[13] = 'SUBOPTIMAL'
status_dict[14] = 'IN_PROGRESS'

row_status_dict = {}
row_status_dict[-1] = 'non-basic'
row_status_dict[ 0] = 'basic'

col_status_dict = {}
col_status_dict[-3] = 'super-basic'
col_status_dict[-2] = 'non-basic at upper bound'
col_status_dict[-1] = 'non-basic at lower bound'
col_status_dict[ 0] = 'basic'

cdef class GurobiModel:
	"""
	Wraps the Gurobi Model
	"""
	
	# GurobiEnvironment and Model
	cdef cgurobi.GRBenv* env
	cdef cgurobi.GRBmodel* model
	cdef int n, m
	
	def __init__(self, int numvars, int_par_method = None, int_par_outputflag = None, dbl_par_walltime = None):
		# allocate gurobi environment
		self.n = numvars
		self.m = -1
		cdef int error = cgurobi.GRBloadenv(&self.env, NULL)
		if error != 0 or self.env == NULL:
			raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
		
		# for some reason those parameters have to be set *before* creating the model
		if int_par_outputflag != None:
			self.int_par_outputflag = int_par_outputflag
		if int_par_method != None:
			self.int_par_method     = int_par_method
		if dbl_par_walltime != None:
			self.dbl_par_walltime   = dbl_par_walltime
		
		# allocate the gurobi model
		error = cgurobi.GRBnewmodel( self.env, &self.model, NULL,
                                     numvars,
                                     NULL, NULL, NULL, NULL, NULL)
		if error != 0:
			raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
		
	def __del__(self):
		# clean up
		cdef int error = cgurobi.GRBfreemodel(self.model)
		if error != 0:
			raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
		cgurobi.GRBfreeenv(self.env)
		
	property int_attr_numvars:
		"""
		Number of Variables in GuRoBi Model
		"""
		def __get__(self):
			cdef int error, numvars
			error = cgurobi.GRBgetintattr(self.model, cgurobi.GRB_INT_ATTR_NUMVARS, &numvars)
			if error != 0:
				raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
			return numvars
		
	property int_attr_cbasis:
		"""
		Column Basis Information
		"""
		def __get__(self):
			cdef int error
			nparr = np.empty([self.m], dtype=np.intc)
			cdef int [:] nparr_view = nparr
			error = cgurobi.GRBgetintattrarray(self.model, cgurobi.GRB_INT_ATTR_CBASIS, 0, nparr.shape[0], &nparr_view[0] if nparr.shape[0] > 0 else NULL)
			if error != 0:
				raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
			return nparr
		
	property int_attr_status:
		"""
		Solution Status of GuRoBi Optimization
		"""
		def __get__(self):
			cdef int error, status
			error = cgurobi.GRBgetintattr(self.model, cgurobi.GRB_INT_ATTR_STATUS, &status)
			if error != 0:
				raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
			return status
		
	property int_attr_vbasis:
		"""
		Variable Basis Information
		"""
		def __get__(self):
			cdef int error
			nparr = np.empty([self.int_attr_numvars], dtype=np.intc)
			cdef int [:] nparr_view = nparr
			error = cgurobi.GRBgetintattrarray(self.model, cgurobi.GRB_INT_ATTR_VBASIS, 0, nparr.shape[0], &nparr_view[0] if nparr.shape[0] > 0 else NULL)
			if error != 0:
				raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
			return nparr

	property int_par_outputflag:
		"""
		GuRoBi Output Flag.
		"""
		def __set__(self, int value):
			cdef int error = cgurobi.GRBsetintparam(self.env, cgurobi.GRB_INT_PAR_OUTPUTFLAG, value)
			if error != 0:
				raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
			
	property int_par_method:
		"""
		GuRoBi Method.
		"""
		def __set__(self, int value):
			cdef int error = cgurobi.GRBsetintparam(self.env, cgurobi.GRB_INT_PAR_METHOD, value)
			if error != 0:
				raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
			
	property dbl_par_walltime:
		"""
		Walltime for GuRoBi Computation
		"""
		def __set__(self, double value):
			cdef int error = cgurobi.GRBsetdblparam(self.env, cgurobi.GRB_DBL_PAR_TIMELIMIT, value)
			if error != 0:
				raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
			
	property dbl_attr_lb:
		"""
		Lower bound on lp variables
		"""
		def __set__(self, double [:] lb):
			cdef int error
			assert lb.shape[0] == self.n
			for ii in range(self.n):
				error = cgurobi.GRBsetdblattrelement(self.model, cgurobi.GRB_DBL_ATTR_LB, ii, lb[ii])
				if error != 0:
					raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
				
	property dbl_attr_ub:
		"""
		Upper bound on lp variables
		"""
		def __set__(self, double [:] ub):
			cdef int error
			assert ub.shape[0] == self.n
			for ii in range(self.n):
				error = cgurobi.GRBsetdblattrelement(self.model, cgurobi.GRB_DBL_ATTR_UB, ii, ub[ii])
				if error != 0:
					raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
				
	property dbl_attr_obj:
		"""
		Gradient vector of lp
		"""
		def __set__(self, double [:] c):
			cdef int error
			assert c.shape[0] == self.n
			for ii in range(self.n):
				error = cgurobi.GRBsetdblattrelement(self.model, cgurobi.GRB_DBL_ATTR_OBJ, ii, c[ii])
				if error != 0:
					raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
				
	property dbl_attr_objval:
		"""
		Objective Value in Solution
		"""
		def __get__(self):
			cdef int error
			cdef double objective
			error = cgurobi.GRBgetdblattr(self.model, cgurobi.GRB_DBL_ATTR_OBJVAL, &objective)
			if error != 0:
				raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
			return objective
		
	property dbl_attr_pi:
		"""
		Dual Solution Vector
		"""
		def __get__(self):
			cdef int error
			nparr = np.empty([self.m])
			cdef double [:] nparr_view = nparr
			error = cgurobi.GRBgetdblattrarray(self.model, cgurobi.GRB_DBL_ATTR_PI, 0, nparr.shape[0], &nparr_view[0] if nparr.shape[0] > 0 else NULL)
			if error != 0:
				raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
			return nparr
		
	property dbl_attr_slack:
		"""
		Row Slack Vector
		"""
		def __get__(self):
			cdef int error
			nparr = np.empty([self.m])
			cdef double [:] nparr_view = nparr
			error = cgurobi.GRBgetdblattrarray(self.model, cgurobi.GRB_DBL_ATTR_SLACK, 0, nparr.shape[0], &nparr_view[0] if nparr.shape[0] > 0 else NULL)
			if error != 0:
				raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
			return nparr
		
	property dbl_attr_x:
		"""
		Primal Solution Vector
		"""
		def __get__(self):
			cdef int error
			nparr = np.empty([self.int_attr_numvars])
			cdef double [:] nparr_view = nparr
			error = cgurobi.GRBgetdblattrarray(self.model, cgurobi.GRB_DBL_ATTR_X, 0, nparr.shape[0], &nparr_view[0] if nparr.shape[0] > 0 else NULL)
			if error != 0:
				raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
			return nparr
				
	def get_str_attr_varname(self, ii):
		"""
		Gets the variable name of the ii-th variable
		"""
		cdef char *varname
		cdef int error
		error = cgurobi.GRBgetstrattrelement(self.model, cgurobi.GRB_STR_ATTR_VARNAME, ii, &varname)
		if error != 0:
			raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
		cdef bytes pyvarname = varname
		return pyvarname
				
	def addrangeconstrs(self, m, nnz, int[:] indptr, int[:] indices, double[:] data, double[:] Lb, double[:] Ub):
		"""
		adds ranged constraints of the form Lb <= coeff*x <= Ub
		nnz, indptr, indices, data describe the matrix coeff (notation and meaning matches scipy.sparse.csr_matrix)
		"""
		self.m = m
		assert Lb.shape[0] == m
		assert Ub.shape[0] == m
		cdef int error = cgurobi.GRBaddrangeconstrs(self.model, m, nnz,
                                           &indptr[0] if indptr.shape[0] > 0 else NULL,
                                           &indices[0] if indices.shape[0] > 0 else NULL,
                                           &data[0] if data.shape[0] > 0 else NULL,
                                           &Lb[0] if m > 0 else NULL, &Ub[0] if m > 0 else NULL, NULL)
		if error != 0:
			raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
		
	def updatemodel(self):
		"""
		updates internal gurobi model with the supplied data
		"""
		cdef int error = cgurobi.GRBupdatemodel(self.model)
		if error != 0:
			raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
		
	def write(self, filename):
		"""
		writes the internal gurobi model to file
		"""
		cdef int error = cgurobi.GRBwrite(self.model, <char *> filename)
		if error != 0:
			raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))
		
	def optimize(self):
		"""
		start GuRoBi optimization algorithm
		"""
		cdef int error = cgurobi.GRBoptimize(self.model)
		if error != 0:
			raise Exception('GuRoBi Error %s: %s' % (error, cgurobi.GRBgeterrormsg(self.env)))

	
def lpsolve(double [:] c, coeff, double[:] lb, double[:] ub, double[:] Lb, double[:] Ub, int verbosity = 0, double walltime = 1e100, write_model=False):
	"""
	Convenience Interface to GuRoBi that solves the LP
	
	min <c,x>
	    s.t. Lb <= coeff*x <= Ub
	    s.t. lb <= x <= ub
	    
	Computation will be stopped after walltime.
	Solution will be returned if any, GuRoBi Model will be destroyed afterwards.
	If you want to employ hotstarts, use the GurobiModel class directly.
	
	Returns a dictionary with solution information
	"""
	if isinstance(coeff, scipy.sparse.csr_matrix):
		A = coeff
	else:
		A = scipy.sparse.csr_matrix(coeff)
	
	cdef double[:] data    = A.data
	cdef int[:]    indices = A.indices
	cdef int[:]    indptr  = A.indptr
	
	cdef int n = A.shape[1]
	cdef int m = A.shape[0]
	
	assert c.shape[0]  == n
	assert lb.shape[0] == n
	assert ub.shape[0] == n
	assert Lb.shape[0] == m
	assert Ub.shape[0] == m
	
	# set up solution status dictionary
	sol = {}
	sol['status'] = -1
	
	# get GuRoBi environment and model
	# use dual simplex (method = 1), set verbosity and walltime
	model = GurobiModel(n, int_par_outputflag = 0 if verbosity == 0 else 1, int_par_method = 1, dbl_par_walltime = walltime)
	
	# set variables, variable bounds and gradient
	model.dbl_attr_lb = lb
	model.dbl_attr_ub = ub
	model.dbl_attr_obj = c
	
	# set constraints and constraint bounds
	cdef int nnz = A.nnz
	model.addrangeconstrs(m, nnz, indptr, indices, data, Lb, Ub)
	
	# update GuRoBi model to take the supplied data
	model.updatemodel()
	
	# with range constraints:
	# for every non-trivial range constraint Gurobi adds another internal variable
    # make sure to track them. this is not efficient at all!
	sol['rangevariable'] = {}
	for ii in range(n, model.int_attr_numvars):
		sol['rangevariable'][int(model.get_str_attr_varname(ii)[3:])] = ii
	
	# if write_model != False write out lp file
	if write_model != False:
		model.updatemodel()
		model.write(write_model)
		
	# solve
	model.optimize()
	
	sol['status'] = model.int_attr_status
	
	# return solution status and solution information if any
	if sol['status'] == 2:
		sol['objective']  = model.dbl_attr_objval
		sol['primal']     = model.dbl_attr_x
		sol['dual']       = model.dbl_attr_pi
		sol['col_status'] = model.int_attr_vbasis
		sol['row_status'] = model.int_attr_cbasis
		sol['row_slack']  = model.dbl_attr_slack

	return sol

def status_info(status):
	return status_dict[status]
    
def row_status_info(status):
	"""
	translates row status into human readable form
	"""
	if isinstance(status, list) or isinstance(status, np.ndarray):
		return [ row_status_dict[litem] for litem in status ]
	else:
		return row_status_dict[status] 

def col_status_info(status):
	"""
	translates column status into human readable form
	"""
	if isinstance(status, list) or isinstance(status, np.ndarray):
		return [ col_status_dict[litem] for litem in status ]
	else:
		return col_status_dict[status] 
   
