#! /usr/bin/env python
# -*- coding: utf-8 -*-

# This software is licensed according to the pySLEQP academic license
# see LICENSE


import os, sys, codecs, locale
import scipy, scipy.sparse
import numpy as np
import eqpsolve
import saddlepointsolve
import time
import traceback

class SLEQP():
    """
    Solves
        min obj(x)
        s.t. c_l <= c(x) <= c_u
             xl <= x <= xu
             
    with an SLEQP method as described in 
        F. Lenders, C. Kirches, H.G. Bock
        pySLEQP: A Sequential Linear Quadratic Programming Method Implemented in Python

        R.H. Byrd, N.I.M. Gould, J. Nocedal, R.A. Waltz
        An algorithm for nonliner optimization using linear programming and equality constraints
    """
    
    def __init__(self, **kwargs):
        """
        Initialize problem
        the following variables must be set using set_problem:
        
            n:                  number of variables
        
            m:                  number of constraints
            
            inequalities:       indices of inequalities
            
            firstorder:         function of (x, userdata) that returns a tuple (objective(x), grad objective(x), cons(x), jac_cons(x))
            
            hess_lgn_prod:      function of (x,s,l,d,userdata) that evaluates H*d where H is the hessian s*f''(x) + sum lambda_i c_i''(x)
            
            hess_lgn_prod_re:   function of (x,s,l,d,userdata) that evaluates  a product of the previous hessian with d. x,s,l are the same
                                as in the previous run but given for convenience
                                
            x:                  current point
            
            xl:                 lower bounds on x
            
            xu:                 upper bounds on x
            
            verbosity:          default: 0, higher verbosity results in more output.
                                even verbosity: do not print copyright message and statistics
                                odd  verbosity:        print copyright message and statistics
                                verbosity that determines intermediate output level: ceil(verbosity/2)
            
            maxiter:            default: 1000, maximum number of iterations
        """
        self.stdoutset = False                                  # flag that indicates if stdout has been set
        self.userdata  = None                                   # user data
        self.n = 0                                              # number of variables
        self.m = 0                                              # number of constraints
        self.m_eq = 0                                           # number of equality constraints
        self.m_ineq = 0                                         # number of inequality constraints
        self.m_active = 0                                       # number of active constraints at current point
        self.m_active_simple = 0                                # number of active constraints at current point including simple bounds
        self.equalities = None                                  # indices of constraints that are equalities
        self.inequalities = None                                # indices of constraints that are inequalities
        self.defacto_simple_eq = None                           # indices of variables where lower bound == uper bound. Don't reset their lagrange multiplier to zero if negative
        self.scale_obj = None                                   # objective scaling factor
        self.scale_obj_inv = None                               # inverse objective scaling factor
        self.scale_var = None                                   # matrix D such that x_scale = D*x
        self.scale_var_inv = None                               # inverse scaling matrix
        self.scale_cons = None                                  # scaling matrix for constraints
        self.scale_cons_inv = None                              # invers scaling matrix
        self.x = None                                           # current point
        self.x_scale = None                                     # scaled current point
        self.x_t = None                                         # trial point
        self.x_t_scale = None                                   # scaled trial point
        self.xl = None                                          # simple lower bounds
        self.xl_scale = None                                    # simple lower bounds scaled
        self.xu = None                                          # simple upper bounds
        self.xu_scale = None                                    # simple upper bounds scaled
        self.cl = None                                          # simple lower bounds
        self.cl_scale = None                                    # simple lower bounds scaled
        self.cu = None                                          # simple upper bounds
        self.cu_scale = None                                    # simple upper bounds scaled
        self.d_c_scale = None                                   # cauchy direction determined by LP
        self.d_ce_scale = None                                  # direction from Cauchy Point to EQP Point
        self.d_n_scale = None                                   # scaled least squares to jacobian_active * d_n = constraints_active
        self.d_t_scale = None                                   # trial direction
        self.cauchy_full = None                                 # flag that indicates if we have taken full LP step as Cauchy step
        self.hess_lgn_prod_c_scale = None                       # product of hessian of lagrangian with d_c
        self.hess_lgn_prod_ce_scale = None                      # product of hessian of lagrangian with d_ce
        self.multiplier_scale = None                            # least squares estimation of lagrange multipliers
        self.multiplier_simple_scale = None                     # least squares estimation of lagrange multipliers for simple bounds
        self.multiplier_eqp_scale = None                        # modified lagrange multipliers used for hessian in EQP subproblem
        self.obj = None                                         # objective value at x
        self.obj_scale = None                                   # scaled objective value at x
        self.obj_t = None                                       # objective value at trial point
        self.obj_t_scale = None                                 # scaled objective value at trial point
        self.merit_scale = None                                 # merit function value at x
        self.merit_t_scale = None                               # merit function value at trial point
        self.grad = None                                        # gradient of objective at x
        self.grad_scale = None                                  # scaled gradient of objective at x
        self.grad_t = None                                      # gradient of objective at trial point
        self.grad_t_scale = None                                # scaled gradient of objective at trial point
        self.cons = None                                        # constraints at x 
        self.cons_scale = None                                  # scaled constraints at x
        self.cons_t = None                                      # constraints at trial point
        self.cons_t_scale = None                                # scaled constraints at trial point
        self.jac_cons = None                                    # jacobian of constraints at x 
        self.jac_cons_scale = None                              # scaled jacobian of constraints at x
        self.jac_cons_t = None                                  # jacobian of constraints at trial point
        self.jac_cons_t_scale = None                            # scaled jacobian of constraints at trial point
        self.grad_eqp_scale = None                              # modified gradient of constraints used in EQP subproblem
        self.firstorder = None                                  # function that evaluates first order information
        self.hess_lgn_prod = None                               # function of (x, s, l,d) that evaluates products of the (s-1)*f''(x) + hessian at (x,l) with a vector d
        self.hess_lgn_prod_re = None                            # function of d that evaluates products of the hessian at the last evaluated point with a vector d
        self.callback = None                                    # callback function of x
        self.active_var_lower = None                            # active lower bounds
        self.active_var_upper = None                            # active upper bounds
        self.active_cons_lower = None                           # active lower bounds
        self.active_cons_upper = None                           # active upper bounds
        self.jac_cons_active_scale = None                       # jacobian of active constraints
        self.ared = None                                        # actual reduction
        self.pred = None                                        # predicted reduction
        self.tr_ratio = None                                    # trust region ratio ared/pred
        self.slacks = None                                      # matrix of slack constraints
        self.statistics = {}                                    # structure for statistics
        self.statistics['iter'] = 0                             # number of iterations
        self.statistics['nsuccess'] = 0                         # number of succesful steps
        self.statistics['nfail'] = 0                            # number of rejected steps
        self.statistics['nobj'] = 0                             # number of objective and gradient evaluations
        self.statistics['ncons'] = 0                            # number of constraints and gradient evaluations
        self.statistics['nhessv'] = 0                           # number of vectorproducts with hessian
        self.statistics['lp_determination'] = 0.                # time spent in LP function
        self.statistics['eval'] = 0.                            # time spent for evaluation of objective, gradient, constraints, gradient, hessian
        self.statistics['active_trial_setup'] = 0.              # time spent during setup of trial active constraints
        self.statistics['saddlepoint_setup'] = 0.               # time spent for setup of saddlepoint matrix
        self.statistics['saddlepoint_factorization'] = 0.       # time spent for setup of saddlepoint matrix and factorization
        self.statistics['auxiliary'] = 0.                       # time spent for computation of directions other than eqp direction
        self.statistics['direction_eqp'] = 0.                   # time spent for computation of eqp direction
        self.statistics['direction_n'] = 0.                     # time spent for computation of n direction
        self.statistics['direction_cauchy'] = 0.                # time spent for computation of cauchy direction
        self.statistics['direction_trial'] = 0.                 # time spent for computation of cauchy direction
        self.statistics['direction_soc'] = 0.                   # time spent for computation of soc direction
        self.statistics['alpha_lp_nf'] = 0                      # number of iterations where alpha_lp was not full
        self.statistics['alpha_qp_nf'] = 0                      # number of iterations where alpha_lp was not full
        self.statistics['multiplier'] = 0.                      # multiplier computation
        self.nhess = 0                                          # counter used in eqp step determination
        self.lasttime = 0.                                      # time used for statistics computation
        self.terminate = -1                                     # termination status
        self.terminatedict = {}                                 # dictionary with termination status
                                                                # nonnegative return code: good reason to terminate
                                                                #    negative return code: solver problem
        self.terminatedict[0]     = 'Locally optimal solution found.'
        self.terminatedict[-1]    = 'Running.'
        # optimal
        # limits reached
        self.terminatedict[-2]   = 'Iteration limit reached.'
        self.terminatedict[-3]   = 'Walltime limit reached.'
        # feasible approximation
        self.terminatedict[10]   = 'Near optimum: primal feasible, close to optimality.'
        self.terminatedict[11]   = 'Primal feasible, step attempted smaller than chtol in maximum norm.'
        self.terminatedict[12]   = 'Primal feasible, optimality could not achieved. Scaling? Degeneracy? Ill-Conditioning.'
        # infeasible termination
        self.terminatedict[20]   = 'Locally infeasible solution found.'
        self.terminatedict[21]   = 'Primal infeasible, step attempted smaller than chtol in maximum norm.'
        self.terminatedict[22]   = 'Primal infeasible, optimality could not achieved. Scaling? Degeneracy? Ill-conditioning?'
        self.terminatedict[23]   = 'Variable bounds infeasible.'
        self.terminatedict[24]   = 'Constraint bounds infeasible.'
        # unbounded
        self.terminatedict[30]   = 'Unbounded from below?'
        # exceptions
        self.terminatedict[-200] = 'Unknown Exception occured'
        # exceptions in algorithm
        self.terminatedict[-201] = 'LP Trust Region too small.'
        self.terminatedict[-202] = 'QP Trust Region too small.'
        self.terminatedict[-203] = 'Predicted Reduction not positive.'
        self.terminatedict[-204] = 'Could not find starting point that satisfies simple bounds.'
        self.terminatedict[-205] = 'NaN found in LP Step.'
        self.terminatedict[-206] = 'NaN found in Norm of LP Step.'
        self.terminatedict[-207] = 'Trust Region Ratio non-numeric (Infinite or NaN).'
        # exceptions from LP Solver || these are - (220 + return code of GuRoBi)
        self.terminatedict[-221] = 'LP Solver Error: Unable to provide Solution information.'
        self.terminatedict[-223] = 'LP Solver Error: LP infeasible.'
        self.terminatedict[-224] = 'LP Solver Error: LP either infeasible or unbounded.'
        self.terminatedict[-225] = 'LP Solver Error: LP unbounded.'
        self.terminatedict[-226] = 'LP Solver Error: LP objective worse than cutoff parameter.'
        self.terminatedict[-227] = 'LP Solver Error: Simplex Iteration Limit reached.'
        self.terminatedict[-229] = 'LP Solver Error: Time Limit exceeded.'
        self.terminatedict[-231] = 'LP Solver Error: Solution process interrupted by user.'
        self.terminatedict[-232] = 'LP Solver Error: Unrecoverable numerical difficulties.'
        self.terminatedict[-233] = 'LP Solver Error: Unable to find optimal solution within tolerances, only sub-optimal solution available.'
        self.terminatedict[-234] = 'LP Solver Error: Optimization still in progress.'
        self.terminatedict[-235] = 'LP Solver Error: Unspecified error occured.'
        self.terminatedict[-236] = 'LP Solver Error: Optimization stopped due to cycling.'
        # exceptions from Linear Algebra || these are - (250 + return code of MA57)
        self.terminatedict[-239] = 'Linear Algebra Error: Iterative Refinement has failed to converge.'
        self.terminatedict[-240] = 'Linear Algebra Error: MeTiS not available.'
        self.terminatedict[-241] = 'Linear Algebra Error: The Array PERM does not hold a permutation.'
        self.terminatedict[-242] = 'Linear Algebra Error: The integer array that holds data for the factors needs to be bigger than CONTROL%%MAXLIW.'
        self.terminatedict[-243] = 'Linear Algebra Error: The real array that holds data for the factors needs to be bigger than CONTROL%%MAXLA.'
        self.terminatedict[-244] = 'Linear Algebra Error: A change in the sign of pivots has been detected.'
        self.terminatedict[-245] = 'Linear Algebra Error: Zero pivot detected.'
        self.terminatedict[-247] = 'Linear Algebra Error: Failure of an allocate or deallocate statement.'
        self.terminatedict[-248] = 'Linear Algebra Error: Value of MATRIX%%NE out of range.'
        self.terminatedict[-249] = 'Linear Algebra Error: Value of MATRIX%%N out of range.'
        self.terminatedict[-254] = 'Linear Algebra Error: Matrix is rank deficient.'
        self.terminatedict[-255] = 'Linear Algebra Error: Pivots have different signs.'
        # exceptions from CG Solver || these are - (290 + return code GLTR)
        self.terminatedict[-246] = 'CG Error: The obtained function value is lower than f_min.'
        self.terminatedict[-260] = 'CG Error: The constraint boundary has been encountered when the input value of control%%stheihaug_toint was set .TRUE.. The solution is unlikely to have achieved the required accuracy.'
        self.terminatedict[-272] = 'CG Error: More than control%%itmax iterations have been performed without obtaining convergence.'
        self.terminatedict[-275] = 'CG Error: The matrix M appears not to be positive definite.'
        self.terminatedict[-277] = 'CG Error: Non-numeric value (Infinite or NaN) in computation of hessian vector product.'
        self.terminatedict[-278] = 'CG Error: Non-numeric value (Infinite or NaN) in computation of projection.'
        self.terminatedict[-287] = 'CG Error: One of the restrictions n > 0 or radius > 0 has been violated.'
        self.terminatedict[-288] = 'CG Error: A deallocation error occured.'
        self.terminatedict[-289] = 'CG Error: An allocation error occured.'
        # exceptions from User Routines
        self.terminatedict[-290] = 'Eval Error: Non-numeric value (Infinite or NaN) in Objective'
        self.terminatedict[-291] = 'Eval Error: Non-numeric value (Infinite or NaN) in Constraints'
        self.terminatedict[-292] = 'Eval Error: Non-numeric value (Infinite or NaN) in Gradient of Objective'
        self.terminatedict[-293] = 'Eval Error: Non-numeric value (Infinite or NaN) in Jacobian of Constraints'
        self.terminatedict[-294] = 'Eval Error: Non-numeric value (Infinite or NaN) in Hessian Vector Product'
        self.terminatedict[-295] = 'Eval Error: Non-numeric value (Infinite or NaN) in Starting Point'
        
        self.opterr = 1.0                                       # optimality error
        self.feaserr = 1.0                                      # feasibility error
        self.optscale = 1.0                                     # scaling factor tau2
        self.feasscale = 1.0                                    # scaling factor tau1
        self.feasscale_init = 1.0                               # constant part of scaling factor tau1
        self.opttol = 1e-6                                      # relative termination tolerance for opterr
        self.feastol = 1e-6                                     # relative termination tolerance for feaserr
        self.chtol = 1e-16                                      # request that || d_t ||_inf > chtol
        self.numrange = 1e20                                    # numerical range, absolute values above of this will be considered as infinite
        self.objrange = 1e20                                    # objective range. if objective values <= -objrange are found, the problem is considered as unbounded
        self.alphamin = 1e-20                                   # minimum stepsize in linesearch
        self.opttol_abs = 0.0                                   # absolute termination tolerance for opterr
        self.feastol_abs = 0.0                                  # absolute termination tolerance for feaserr
        self.tr_eqp = 1                                         # EQP trust region radius
        self.tr_eqp_null = None                                 # trust region radius of zero space EQP
        self.tr_lp = None                                       # LP trust region radius, will be set to .8 * tr_eqp / sqrt(n), cf Byrd 2003
        self.tr_acceptance = 1e-8                               # if tr_ratio >= tr_acceptance, the trial step will be accepted
        self.penalty = 1e1                                      # penalty parameter as given in Byrd 2003
        self.penaltytolarge = 0                                 # number of iterates the penalty parameter has not been adjusted
        self.tol_1 = 1e-8                                       # tolerance 1 in penalty update, cf Byrd 2003
        self.tol_2 = 1e-8                                       # tolerance 2 in penalty update, cf Byrd 2003
        self.tol_cmp_rel = 1e-6                                 # relative tolerance that constraints have to satisfy
        self.tol_cmp_abs = 1e-10                                # absolute tolerance that constraints have to satisfy
        self.epsilon = 0.1                                      # epsilon for backtracking line search in penalty update, cf Byrd 2003
        self.eta_lp = 0.1                                       # not specified in Byrd 2003 except 0 < eta < 1, used in backtracking linesearch for cauchy point
        self.eta_eqp = 1e-4                                     # used in backtracking linesearch for eqp point
        self.tau = 0.5                                          # not specified in Byrd 2003 except 0 < tau < 1, step reduction in linesearch
        self.verbosity = 0                                      # verbosity paramter
        self.copyrightandstat = False                           # = given verbosity %2
        self.logfile = 0                                        # write logfile
        self.compute_cond = False                               # compute condition
        self.maxiter = 1000                                     # maximum number of iterations
        self.walltime = 60*60                                   # maximum solution time
        self.zerotime = 0                                       # system time at which solution process startet
        self.preconditioner = 1.                                # preconditioner scalar
        self.stepaccepted = True                                # flag that indicates if step is accepted
        self.gltrproblem40 = False                              # flag that indicates if GLTR Warning -40 occured
        self.gltrproblem30 = False                              # flag that indicates if GLTR Warning -30 occured
        self.socattempted = False                               # flag that indicates if SOC was attempted
        self.predzero = False                                   # flag that indicates if predicted reduction was zero
        self.prednotpositivesoc = False                         # flag that indicates if predicted reduction was non-positive after soc step
        self.lpsolver = 'gurobi'                                # lp solver
        self.lpsolvelib = None                                  # library to lpsolver
        self.real_lower = None                                  # real lower variable inequalities in current iteration
        self.real_upper = None                                  # real upper variable inequalities in current iteration
        self.lp_objective = None                                # lp data for parametric resolve
        self.lp_cons_lower = None
        self.lp_cons_upper = None
        self.lp_var_lower = None
        self.lp_var_upper = None
        self.set_problem(**kwargs) 
        
    def __del__(self):
        saddlepointsolve.finalize_saddlepoint()
        
    def colour_string(self, string, colour='blue'):
        colourcode = ''
        if colour == 'blue':
            colourcode = u'\033[94m'
        if colour == 'red':
            colourcode = u'\033[91m'
        if colour == 'green':
            colourcode = u'\033[92m'
        return colourcode + string + u'\033[00m'
        
    def neglog10(self, number, format="%6.2f", positive = False):
        if number == 0.:
            dummy = format% 1.0
            length = len(dummy)
            if positive:
                return " ".join(["" for ii in range(length-1)]) + u"-∞"
            else:
                return " ".join(["" for ii in range(length-0)]) + u"∞"
        else:
            return format % np.log10(number) if positive else format % -np.log10(number)
        
           
    def set_problem(self, **kwargs):
        if 'n' in kwargs:
            self.n = kwargs['n']
            self.d_c_scale = np.empty([self.n])
            self.d_ce_scale = np.empty([self.n])
            self.tr_lp = .8*self.tr_eqp/np.sqrt(self.n)
            self.grad_eqp_scale = np.empty([self.n])
        if 'm' in kwargs:
            self.m = kwargs['m']
            self.multiplier_eqp_scale = np.empty([self.m])
        if 'hess_lgn_prod' in kwargs:
            self.hess_lgn_prod = kwargs['hess_lgn_prod']
        if 'hess_lgn_prod_re' in kwargs:
            self.hess_lgn_prod_re = kwargs['hess_lgn_prod_re']
        if 'firstorder' in kwargs:
            self.firstorder = kwargs['firstorder']
        if 'x' in kwargs:
            self.x = kwargs['x']
        if 'xl' in kwargs:
            self.xl = kwargs['xl']
        if 'xu' in kwargs:
            self.xu = kwargs['xu']
        if 'cl' in kwargs:
            self.cl = kwargs['cl']
        if 'cu' in kwargs:
            self.cu = kwargs['cu']
        if 'verbosity' in kwargs:
            self.verbosity = int(np.ceil(.5*kwargs['verbosity']))
            self.copyrightandstat = kwargs['verbosity']%2 == 1
        if 'compute_cond' in kwargs:
            self.compute_cond = kwargs['compute_cond']
        if 'maxiter' in kwargs:
            self.maxiter = kwargs['maxiter']
        if 'callback' in kwargs:
            self.callback = kwargs['callback']
        if 'log' in kwargs:
            self.logfile = kwargs['log']
        if 'lpsolver' in kwargs:
            self.lpsolver = kwargs['lpsolver']
        if 'stdout' in kwargs:
            sys.stdout = kwargs['stdout']
            self.stdoutset = True
        if 'userdata' in kwargs:
            self.userdata = kwargs['userdata']
            
    def feasibility_measure(self, d):
        if self.m == 0:
            return 0.
        else:
            residual = self.jac_cons_scale.dot(d) + self.cons_scale
            residual_lower = (residual - self.cl_scale).clip(max=0.) # those >= 0 are fine...
            residual_upper = (self.cu_scale - residual).clip(max=0.) # ...
            return ( np.linalg.norm( residual_lower, 1 ) + np.linalg.norm( residual_upper, 1) )/self.m
        
    def check_lp_solution_status(self, status):
        if self.lpsolver == 'gurobi':
            if status == 2:
                return 0
            else:
                self.terminate = -220 - status
                return -1
        if self.lpsolver == 'soplex':
            if status == 1:
                return 0
            else:
                if status in [-13, -12, -11, -10, -9, -3, -2, 0]: # ERROR, NO_RATIOTESTER, NO_PRICER, NO_SOLVER, NOT_INIT, NO_PROBLEM, REGULAR
                    self.terminate = -235
                if status == -8: # CYCLING
                    self.terminate = -236
                if status == -7: # ABORT_TIME
                    self.terminate = -229
                if status == -6: # ABORT_ITER
                    self.terminate = -227
                if status == -5: # ABORT_VALUE
                    self.terminate = -226
                if status == -4: # SINGULAR
                    self.terminate = -232
                if status == -1: # RUNNING
                    self.terminate = -234
                if status == 0:
                    self.terminate = -221
                if status == 2: # UNBOUNDED
                    self.terminate = -225
                if status == 3: # INFEASIBLE
                    self.terminate = -223
                if status == 4: # INForUNBD
                    self.terminate = -224
                return -1
            
    def get_active_set(self, solution=None, ws=None):
        # read active set
        self.active_cons_lower = []
        self.active_cons_upper = []
        self.active_var_lower  = []
        self.active_var_upper  = []
        
        if ws == None:
            if self.lpsolver in ['gurobi']:
                for ii in range(self.m):
                    lower_bd_constraint = True
                    upper_bd_constraint = True
                    if ii in solution['rangevariable']:
                        if solution['col_status'][solution['rangevariable'][ii]] == 0:
                            continue
                        else:
                            lower_bd_constraint = solution['col_status'][solution['rangevariable'][ii]] == -2
                            upper_bd_constraint = solution['col_status'][solution['rangevariable'][ii]] == -1
                    if solution['row_status'][ii] != 0:
                        # equality case
                        if self.cl[ii] == self.cu[ii]:
                            if solution['col_status'][self.n + ii] != 0 and solution['col_status'][self.n + self.m + ii] != 0:
                                self.active_cons_lower.append(ii)
                        else:
                            if solution['col_status'][self.n + ii] != 0 and self.cl_scale[ii] > - self.numrange and lower_bd_constraint:
                                self.active_cons_lower.append(ii)
                            elif solution['col_status'][self.n + self.m + ii] != 0 and self.cu_scale[ii] < self.numrange and upper_bd_constraint:
                                self.active_cons_upper.append(ii)
                for ii in range(self.n):
                    if self.real_lower[ii] >= - self.tr_lp and solution['col_status'][ii] != 0:
                        self.active_var_lower.append(ii)
                    if self.real_upper[ii] <=   self.tr_lp and ii not in self.active_var_lower and solution['col_status'][ii] != 0:
                        self.active_var_upper.append(ii)
                        
            if self.lpsolver in ['soplex']:
                for ii in range(self.m):
                    if solution['row_status'][ii] == 4:
                        continue
                    lower_bd_constraint = solution['row_status'][ii] == 1 or (self.lp_cons_lower[ii] == 0. and solution['row_status'][ii] == 2)
                    upper_bd_constraint = solution['row_status'][ii] == 0 or (self.lp_cons_upper[ii] == 0. and solution['row_status'][ii] == 2)
                    if self.cl[ii] == self.cu[ii]:
                        if solution['col_status'][self.n + ii] != 4 and solution['col_status'][self.n + self.m + ii] != 4:
                            self.active_cons_lower.append(ii)
                    else:
                        if solution['col_status'][self.n + ii] != 4 and self.cl_scale[ii] > - self.numrange and lower_bd_constraint:
                            self.active_cons_lower.append(ii)                        
                        elif solution['col_status'][self.n + self.m + ii] != 4 and self.cu_scale[ii] < self.numrange and upper_bd_constraint:
                            self.active_cons_upper.append(ii)
                for ii in range(self.n):
                    if   self.real_lower[ii] >= -self.tr_lp and solution['col_status'][ii] != 4:
                        self.active_var_lower.append(ii)
                    elif self.real_upper[ii] <=  self.tr_lp and solution['col_status'][ii] != 4:
                        self.active_var_upper.append(ii)
                    
        self.active_cons_lower = np.array(self.active_cons_lower, dtype=np.int)
        self.active_cons_upper = np.array(self.active_cons_upper, dtype=np.int)
        self.active_var_lower  = np.array(self.active_var_lower,  dtype=np.int)
        self.active_var_upper  = np.array(self.active_var_upper,  dtype=np.int)
        
        if self.logfile:
            self.file.write('LP Trust Region:   %e\n' % self.tr_lp)
            self.file.write('Penalty Parameter: %e\n' % self.penalty)
            self.file.write('LP Direction:      '); np.savetxt(self.file, self.d_c_scale, newline=" "); self.file.write('\n')
            self.file.write('Active Lower C:    '); np.savetxt(self.file, self.active_cons_lower, newline=" ", fmt='%u'); self.file.write('\n')
            self.file.write('Active Upper C:    '); np.savetxt(self.file, self.active_cons_upper, newline=" ", fmt='%u'); self.file.write('\n')
            self.file.write('Active Lower V:    '); np.savetxt(self.file, self.active_var_lower, newline=" ", fmt='%u'); self.file.write('\n')
            self.file.write('Active Upper V:    '); np.savetxt(self.file, self.active_var_upper, newline=" ", fmt='%u'); self.file.write('\n')
            
    def solve_lp(self, simple_infeasible=False):
        """
        Solve LP :
            linearized l1 merit function as objective using the given penalty parameter,
            linearized constraints and simple bounds as constraints
            subject to trust region constraint self.tr_lp
        
        The solution will be used to determine the active set and the Cauchy point.
        The penalty parameter will eventually be updated to ensure feasibility
        """
        
        # form constraint matrix of augmented problem
        # first n variables:           corresponding to n problem variables
        # subsequent n_eq   variables: corresponding to positive slacks of n_eq variables
        # subsequent n_eq   variables: corresponding to negative slacks of n_eq variables
        # subsequent n_ineq variables: corresponding to positive slacks of n_ineq variables
        
        # Damn hstack which can not handle empty matrices!
        if self.m > 0:
            lp_augment_constraints = scipy.sparse.hstack( [self.jac_cons_scale, self.slacks] )
        else:
            lp_augment_constraints = self.jac_cons_scale
       
        self.lp_cons_lower                    = self.cl_scale - self.cons_scale
        self.lp_cons_upper                    = self.cu_scale - self.cons_scale
        
        self.lp_var_lower          = np.zeros([self.n + 2 * self.m])
        self.lp_var_upper          = np.inf * np.ones([self.n + 2 * self.m])
        self.real_lower            = self.xl_scale - self.x_scale
        self.real_upper            = self.xu_scale - self.x_scale
        if simple_infeasible:
            #lp_var_lower[:self.n] = real_lower
            #lp_var_upper[:self.n] = real_upper
            slacklower = ( self.real_lower).clip(0)
            slackupper = (-self.real_upper).clip(0)
            allowed_dist = self.xu_scale - self.xl_scale
            if (slacklower>0).any():
                self.x_scale = self.x_scale + np.minimum(1.1*slacklower, slacklower + .5*allowed_dist)
            if (slackupper>0).any():
                self.x_scale = self.x_scale - np.minimum(1.1*slackupper, slackupper + .5*allowed_dist)
            self.x = self.scale_var_inv.dot(self.x_scale)
            return 0
        else:
            self.lp_var_lower[:self.n] = np.fmax(-self.tr_lp, self.real_lower)
            self.lp_var_upper[:self.n] = np.fmin( self.tr_lp, self.real_upper)
            
        # close logfile since the LP solver should be able to write in it
        if self.verbosity > 1 and self.logfile != False:
            self.file.close()
        
        self.lp_objective          = self.penalty * np.ones([self.n + 2 * self.m])
        self.lp_objective[:self.n] = self.grad_scale
        
        if self.lpsolver in ['gurobi', 'soplex']:
            solution = self.lpsolvelib.lpsolve( self.lp_objective, lp_augment_constraints,
                                                self.lp_var_lower, self.lp_var_upper, self.lp_cons_lower, self.lp_cons_upper,
                                                walltime = self.walltime)
            if self.check_lp_solution_status(solution['status']) < 0:
                return -1
            self.d_c_scale = solution['primal'][:self.n]
                        
        # check infeasibility
        xi_cur = self.feasibility_measure(self.d_c_scale)
       
        # if feasibility measure detects infeasibility, update penalty parameter
        if xi_cur > self.tol_1:
            self.penaltytolarge = 0
            # compute approximation to feasibility measure with infinite penalty parameter
            lp_objective_zero_grad = np.copy(self.lp_objective)
            lp_objective_zero_grad[:self.n] = 0.
            
            if self.lpsolver in ['gurobi', 'soplex']:
                solution2 = self.lpsolvelib.lpsolve( lp_objective_zero_grad, lp_augment_constraints,
                                                self.lp_var_lower, self.lp_var_upper, self.lp_cons_lower, self.lp_cons_upper,
                                                walltime = self.walltime )
                if self.check_lp_solution_status(solution2['status']) < 0:
                    return -1
                d_c_scale2 = solution2['primal'][:self.n]
                
            xi_inf = self.feasibility_measure(d_c_scale2)
           
            # if problem with infinite penalty feasible, increase penalty
            if xi_inf < self.tol_1:
                while xi_cur > self.tol_1 and self.penalty < self.numrange:
                    self.penalty *= 10
                    self.lp_objective[self.n:] = self.penalty
                    
                    if self.lpsolver in ['gurobi', 'soplex']:
                        solution = self.lpsolvelib.lpsolve( self.lp_objective, lp_augment_constraints,
                                                            self.lp_var_lower, self.lp_var_upper, self.lp_cons_lower, self.lp_cons_upper,
                                                            walltime = self.walltime )
                        if self.check_lp_solution_status(solution['status']) < 0:
                            return -1
                        self.d_c_scale = solution['primal'][:self.n]
                        
                    xi_cur = self.feasibility_measure(self.d_c_scale)
                    
            # if xi_cur - xi_inf < tol_2: keep penalty, since no significant progress in feasibility possible
            # otherwise perform backtracking linesearch to ensure significant progress in feasibility
            elif xi_cur - xi_inf >= self.tol_2:
                    xi_test = xi_cur
                    while xi_cur - xi_test < self.epsilon * ( xi_cur - xi_inf) and self.penalty < self.numrange:
                        self.penalty *= 10
                        self.lp_objective[self.n:] = self.penalty
                        
                        if self.lpsolver in ['gurobi', 'soplex']:
                            solution = self.lpsolvelib.lpsolve( self.lp_objective, lp_augment_constraints,
                                                                self.lp_var_lower, self.lp_var_upper, self.lp_cons_lower, self.lp_cons_upper,
                                                                walltime = self.walltime )
                            if self.check_lp_solution_status(solution['status']) < 0:
                                return -1
                            self.d_c_scale = solution['primal'][:self.n]
                           
                        xi_test = self.feasibility_measure(self.d_c_scale)
        
        if self.verbosity > 0:
            print self.neglog10(self.penalty, format="%4.1f", positive=True),
            sys.stdout.flush()
                        
        if np.isnan(self.d_c_scale).any():
            self.terminate = -205
            return -1
        
        if np.isnan(self.penalty):
            self.terminate = -206
            return -1
        
        if self.verbosity > 1 and self.logfile != False:
            self.file = open(self.logfile, 'a')
        
        self.get_active_set(solution if self.lpsolver in ['gurobi', 'soplex'] else p)
        
        self.statistics['lp_determination'] += time.clock() - self.lasttime; self.lasttime = time.clock()
        return 0
    
    def get_next_tr_breakpoint(self, tr):
        spaceup = ( self.real_upper - tr).clip(min=0.)
        spacelo = (-self.real_lower - tr).clip(min=0.)
        minup = np.argmin( ( self.real_upper - tr).clip(min=0.) )
        minlo = np.argmin( (-self.real_lower - tr).clip(min=0.) )
        m = np.inf
        if minup.nonzero()[0].shape[0] > 0:
            m = np.minimum(m, self.real_upper[spaceup.nonzero()[0][np.argmin(spaceup[spaceup.nonzero()])]])
        if minlo.nonzero()[0].shape[0] > 0:
            m = np.minimum(m, -self.real_lower[spacelo.nonzero()[0][np.argmin(spacelo[spacelo.nonzero()])]])
        if m < 0:
            sys.exit()
        if m == np.inf:
            return tr
        else:
            return m
        
    def build_active_constraints_matrix(self, active_cons_lower, active_cons_upper, active_var_lower, active_var_upper):
        """
        Output
        [ Jc[active_cons_lower,:]          ]
        [ Jc[active_cons_upper,:]          ]
        [  e_ii for ii in active_var_lower ]
        [ -e_ii for ii in active_var_upper ]
        """
        len_simple = active_var_lower.shape[0]+active_var_upper.shape[0]
        if len_simple > 0:
            data_simple = np.ones([len_simple])
            data_simple[active_var_lower.shape[0]:] = -1
            col_simple = np.concatenate([active_var_lower, active_var_upper])
            A_simple = scipy.sparse.coo_matrix( (data_simple, (range(data_simple.shape[0]), col_simple)), shape=(len_simple, self.n)).tocsr()
        
        if (active_cons_lower.shape[0] + active_cons_upper.shape[0]) > 0:
            A_act = self.jac_cons_scale[ np.concatenate([active_cons_lower, active_cons_upper]), :]
            if len_simple > 0:
                return scipy.sparse.vstack( [A_act, A_simple], format='csr' )
            else:
                return A_act
        elif len_simple > 0:
            return A_simple
        else:
            return scipy.sparse.csr_matrix( (0, self.n) )

    def setup_active_constraints(self):
        """
        Setup Active Constraints and Jacobian
        """
        self.m_active = self.active_cons_lower.shape[0] + self.active_cons_upper.shape[0]
        self.jac_cons_active_scale = self.build_active_constraints_matrix( self.active_cons_lower, self.active_cons_upper, self.active_var_lower, self.active_var_upper )
        self.m_active_simple = self.m_active + self.active_var_lower.shape[0] + self.active_var_upper.shape[0]
        self.preconditioner = 1.
        self.statistics['saddlepoint_setup'] += time.clock() - self.lasttime; self.lasttime = time.clock()
        status = saddlepointsolve.factor_saddlepoint(self.jac_cons_active_scale.tocoo(), np.abs(1.0/self.preconditioner) )
        self.statistics['saddlepoint_factorization'] += time.clock() - self.lasttime; self.lasttime = time.clock()       
        if status < 0 or status == 4:
            self.terminate = -250 - status
            return -1
        return 0
            
    def merit_eval(self, trial=False):
        if not trial:
            residual_lower = (self.cons_scale - self.cl_scale).clip(max=0.) # those >= 0 are fine...
            residual_upper = (self.cu_scale - self.cons_scale).clip(max=0.) # ...
            self.merit_scale = self.obj_scale + self.penalty * ( np.linalg.norm(residual_lower, 1) + np.linalg.norm(residual_upper, 1) )
        else:
            residual_lower = (self.cons_t_scale - self.cl_scale).clip(max=0.) # those >= 0 are fine...
            residual_upper = (self.cu_scale - self.cons_t_scale).clip(max=0.) # ...
            self.merit_t_scale = self.obj_t_scale + self.penalty * ( np.linalg.norm(residual_lower, 1) + np.linalg.norm(residual_upper, 1) )
            
    def merit_linearization(self, d, trial=False):
        """
        Computes value of linearized merit function
        in Byrd 2003 it is just gradient information
        without a function values which gives scaling
        problems in the line search for alpha_lp.
        Therefor we include the function value as
        in the definition of Byrd 2005
        """
        if not trial:
            ob = self.obj_scale + np.dot(self.grad_scale, d)
            cons = self.cons_scale + self.jac_cons_scale.dot(d)
        else:
            ob = self.obj_t_scale + np.dot(self.grad_t_scale,d)
            cons = self.cons_t_scale + self.jac_cons_t_scale.dot(d)
        residual_lower = (cons - self.cl_scale).clip(max=0.)
        residual_upper = (self.cu_scale - cons).clip(max=0.)
        return ob + self.penalty * ( np.linalg.norm(residual_lower,1) + np.linalg.norm(residual_upper, 1) )
    
    def merit_linearization_grad_prod(self, d_eval, d):
        """
        Computes the directional directive in direction d
        of the linearized merit function at point d_eval
        """
        return 1e8*(self.merit_linearization(d_eval + 1e-8*d) - self.merit_linearization(d_eval))
           
    def direction_cauchy(self, test_reduction=False, complete_parametric=False):
        """
        Compute Cauchy Step as approximate minimizer of a quadratic model
        of the merit function along the direction given by the LP.
        Backtracking line search is used.
        """
        norm_c = np.linalg.norm(self.d_c_scale)
        if np.isnan(norm_c):
            self.terminate = -206
            return -1
        
        alpha_lp = 1.0
        if norm_c > 0:
            alpha_lp = min( 1.0, self.tr_eqp / np.linalg.norm(self.d_c_scale) )
        alpha_lp_max = alpha_lp
        merit_linearized = self.merit_linearization(alpha_lp*self.d_c_scale)
        self.statistics['direction_cauchy'] += time.clock() - self.lasttime; self.lasttime = time.clock()       
        self.statistics['nhessv'] += 1
        self.hess_lgn_prod_c_scale = self.scale_var_inv.dot(self.hess_lgn_prod(self.x, self.scale_obj, -self.scale_cons.dot(self.multiplier_scale), self.scale_var_inv.dot(self.d_c_scale),self.userdata))
        if not np.isfinite(self.hess_lgn_prod_c_scale).all():
            self.terminate = -294
            return
        self.statistics['eval'] += time.clock() - self.lasttime; self.lasttime = time.clock()              
        q_dir_lp = self.hess_lgn_prod_c_scale.dot(self.d_c_scale)
        q = merit_linearized + .5 * alpha_lp * alpha_lp * q_dir_lp
        
        if self.logfile:
            self.file.write('qlp, q, meritlin:  %e, %e, %e\n' % (q_dir_lp, q, merit_linearized) )
            
        if test_reduction:
            if self.merit_scale - q >= self.eta_lp * (self.merit_scale - merit_linearized):
                return 0, self.merit_scale - q
            else:
                return 1, self.merit_scale - q
            
        if not complete_parametric:
            while self.merit_scale - q < self.eta_lp * (self.merit_scale - merit_linearized) and alpha_lp >= self.alphamin:
                alpha_lp = alpha_lp * self.tau
                merit_linearized = self.merit_linearization(alpha_lp*self.d_c_scale)
                q = merit_linearized + .5 * alpha_lp * alpha_lp * q_dir_lp
            if alpha_lp < self.alphamin:
                alpha_lp = 0.
        
        if self.verbosity > 0:
            print self.neglog10(self.tr_lp), 
            print self.neglog10(norm_c),
            print "%7.1e" % alpha_lp if alpha_lp == 1.0 else self.colour_string("%7.1e" % alpha_lp, colour='red'), 
            sys.stdout.flush()
            
        if alpha_lp < 1.0:
            self.statistics['alpha_lp_nf'] += 1
            
        if self.logfile:
            self.file.write('alpha_lp, max, q:     %e, %e, %e\n' % (alpha_lp, alpha_lp_max, q))
        
        self.d_c_scale = alpha_lp * self.d_c_scale
        self.hess_lgn_prod_c_scale *= alpha_lp
        self.cauchy_full = alpha_lp == 1.0
        self.statistics['direction_cauchy'] += time.clock() - self.lasttime; self.lasttime = time.clock()
        return 0
        
    def direction_ce(self):
        """
        Compute direction from EQP point to Cauchy Point.
        Cauchy Point:                           x_c = x + d_c
        EQP Point:                              x_eqp = x + d_n + d_eqp
        direction from Cauchy to EQP Point:     d_ce = x_eqp - x_c
        
        Compute d_eqp as solution of the equality constrained
        zero space problem
        
        min 1/2 <d_eqp, H*d_eqp> + <g_eqp, d_eqp>
        
        s.t. A*d_eqp = 0        (A jacobian of active constraints)
             || d_eqp || <= sqrt( tr_eqp^2 - ||d_n||^2 )
        
        with modified hessian H_eqp and gradient g_eqp,
        d_n is the image space vector.
        """
        
        # obtain eqp direction d_eqp and store it in d_ce
        
        # function that computes the projection step
        self.nhess = 0
        def proj(vector):
            status = saddlepointsolve.solve_saddlepoint_overwrite_s_zero_t(self.n, self.m_active_simple, vector)
            if status < 0:
                self.terminate = -250 - status
        # function that computes the product with the hessian
        def h_prod(vector):
            self.statistics['direction_eqp'] += time.clock() - self.lasttime; self.lasttime = time.clock()       
            self.statistics['nhessv'] += 1
            if self.nhess == 0:
                vector[:] = self.scale_var_inv.dot( self.hess_lgn_prod(self.x, self.scale_obj, -self.scale_cons.dot(self.multiplier_eqp_scale), self.scale_var_inv.dot(vector), self.userdata) )
            else:
                vector[:] = self.scale_var_inv.dot( self.hess_lgn_prod_re(self.x, self.scale_obj, -self.scale_cons.dot(self.multiplier_eqp_scale), self.scale_var_inv.dot(vector), self.userdata) )
            if not np.isfinite(vector).all():
                self.terminate = -294
                return -1
            self.nhess += 1
            self.statistics['eval'] += time.clock() - self.lasttime; self.lasttime = time.clock()
            return 0
            
        self.statistics['direction_eqp'] += time.clock() - self.lasttime; self.lasttime = time.clock()              
        self.d_ce_scale, self.preconditioner, info, num_iter = eqpsolve.eqpsolve(self.n, self.tr_eqp_null, self.grad_eqp_scale, h_prod, proj, self.walltime)
        if info == -40:  # termination code -40 has been introduced in a modified GLTR version to notify if the case <p, Hp> = 0 occures.
            self.gltrproblem40 = True
            self.preconditioner = 1.0
            if self.verbosity > 0:
                print self.colour_string("%4d" % num_iter, colour="blue"),
                sys.stdout.flush()
        elif info == -30:
            self.gltrproblem30 = True
            if self.verbosity > 0:
                print self.colour_string("%4d" % num_iter, colour="red"),
                sys.stdout.flush()
        else:
            if self.verbosity > 0:
                print "%4d" % num_iter,
                sys.stdout.flush()
                
        if self.logfile:
            if info == -40:
                self.file.write("GLTR bailed out with <p, Hp> = 0\n")
            if info == -30:
                self.file.write("GLTR warning possible inaccurate boundary solution\n")
            self.file.write('EQP Solution:      '); np.savetxt(self.file, self.d_ce_scale, newline=" "); self.file.write('\n')
        
        if info < 0 and info not in [-30, -40]: # -30 is only a warning, no error and can appear regardless to the value of steihaug_toint in GLTR
            if info == -13 and self.terminate != -294:
                self.terminate = -290 - info
            
        if self.terminate <= -200:
            return -1
                   
        # compute direction d_ce from Cauchy Point to EQP Point
        # d_ce = d_n + d_eqp - d_c
        self.d_ce_scale = self.d_ce_scale + self.d_n_scale - self.d_c_scale
        
        if self.logfile:
            self.file.write('EQP Direction:     '); np.savetxt(self.file, self.d_ce_scale, newline=" "); self.file.write('\n')
               
        # active simple bounds are forced to be satsisfied in the EQP step
        # hence we have to make sure, that d_ce satisfies them as well
        if self.cauchy_full:
            self.d_ce_scale[self.active_var_lower] = 0
            self.d_ce_scale[self.active_var_upper] = 0
            
        if self.logfile:
            self.file.write('EQP Dir w simple:  '); np.savetxt(self.file, self.d_ce_scale, newline=" "); self.file.write('\n')

        self.statistics['direction_eqp'] += time.clock() - self.lasttime; self.lasttime = time.clock()
        self.statistics['nhessv'] += 1
        self.hess_lgn_prod_ce_scale = self.scale_var_inv.dot(self.hess_lgn_prod(self.x, self.scale_obj, -self.scale_cons.dot(self.multiplier_scale), self.scale_var_inv.dot(self.d_ce_scale), self.userdata))
        if not np.isfinite(self.hess_lgn_prod_ce_scale).all():
            self.terminate = -294
            return -1
        self.statistics['eval'] += time.clock() - self.lasttime; self.lasttime = time.clock()
        
        return 0

    def direction_n(self):
        """
        Compute d_n as the scaled least squares solution to
        jacobian active * d_n = constraints active
        such that || d_n ||_2 <= tr_eqp.
        """
        
        if self.m_active == 0:
            self.d_n_scale = np.zeros(self.n)
            self.tr_eqp_null = self.tr_eqp
            self.statistics['direction_n'] += time.clock() - self.lasttime; self.lasttime = time.clock()                          
            return
        
        rhs_cons = np.empty([self.m_active])
        rhs_cons[:self.active_cons_lower.shape[0]] = self.cons_scale[self.active_cons_lower] - self.cl_scale[self.active_cons_lower]
        rhs_cons[self.active_cons_lower.shape[0]:] = self.cons_scale[self.active_cons_upper] - self.cu_scale[self.active_cons_upper]
        
        rhs = - np.concatenate((rhs_cons,
                               self.x_scale[self.active_var_lower] - self.xl_scale[self.active_var_lower],
                               self.xu_scale[self.active_var_upper] - self.x_scale[self.active_var_upper]))
        self.d_n_scale = saddlepointsolve.solve_saddlepoint_get_s_zero_s(self.n, self.m_active_simple, rhs)
        nd_n = np.linalg.norm(self.d_n_scale)
        
        # ensure that norm of d_n does not exceed tr_eqp
        alpha = 1.
        if nd_n > 0:
            alpha = min(.8*self.tr_eqp/nd_n, 1.)
        self.d_n_scale = alpha * self.d_n_scale
        
        # adjust trust region radius for zero space EQP
        self.tr_eqp_null = np.sqrt( self.tr_eqp * self.tr_eqp - alpha*alpha*nd_n*nd_n )
        self.statistics['direction_n'] += time.clock() - self.lasttime; self.lasttime = time.clock()                          
        
    def divide_clip_zero(self, a, b, cutoff = 1., positive = True):
        """
        Helper Function that can be used to compute min(cutoff, min(a ./ |b|)) where only positive
        or negative values of b are taken
        """
        if positive:
            indices = np.where(b>0.)[0]
            sign = 1.0
        else:
            indices = np.where(b<0.)[0]
            sign = -1.0
        if indices.shape[0] == 0:
            return cutoff
        else:
            with np.errstate(over='ignore'):
                return min(cutoff, min(np.divide(sign*a[indices],b[indices])))
        
    def direction_trial(self):
        """
        Compute trial direction as minimizer of a quadratic model of the penalty function
        on the the line d_c + alpha_eqp d_ce.
        Backtracking line search is used for the determination of the steplength alpha_eqp.
        In Byrd 2003 it is stated that simply decrease of the quadratic model is requested.
        Simple Bounds have to be imposed on the step. d_c satisfies them, d_ce not necessarily.
        """
        
        # restrict steplength so that simple bounds are satisfied and it is not more than 1
        
       alpha_upper_max = self.divide_clip_zero(  self.xu_scale - self.x_scale - self.d_c_scale, self.d_ce_scale, cutoff=1., positive = True )
        alpha_lower_max = self.divide_clip_zero( -self.xl_scale + self.x_scale + self.d_c_scale, self.d_ce_scale, cutoff=1., positive = False )

        if self.verbosity > 0:
            print self.neglog10(self.tr_eqp_null),
            print self.colour_string(self.neglog10(np.linalg.norm(self.d_ce_scale)), colour='red') if (min(alpha_lower_max, alpha_upper_max) <= 1e-6) else self.neglog10(np.linalg.norm(self.d_ce_scale)), 
            sys.stdout.flush()
        
        # if alpha_eqp_max is mildly small
        # project d_ce into simple constraints box
        while alpha_lower_max <= 1e-6:
            indices_obstacle_lo = np.where( self.xl_scale - self.x_scale - self.d_c_scale - self.d_ce_scale > 0 )[0]
            imin = indices_obstacle_lo[np.argmin( np.divide( (self.xl_scale - self.x_scale - self.d_c_scale)[indices_obstacle_lo], self.d_ce_scale[indices_obstacle_lo] ) )]
            if self.logfile:
                self.file.write('projected l index: %d\n' % imin )
            self.d_ce_scale[imin] = (self.xl_scale - self.x_scale - self.d_c_scale)[imin]
            alpha_lower_max = self.divide_clip_zero( -self.xl_scale + self.x_scale + self.d_c_scale, self.d_ce_scale, cutoff=1., positive = False )
        while alpha_upper_max <= 1e-6:
            indices_obstacle_up = np.where( self.xu_scale - self.x_scale - self.d_c_scale - self.d_ce_scale < 0 )[0]
            imin = indices_obstacle_up[np.argmin( np.divide( (self.xu_scale - self.x_scale - self.d_c_scale)[indices_obstacle_up], self.d_ce_scale[indices_obstacle_up] ) )]
            if self.logfile:
                self.file.write('projected l index: %d\n' % imin )
            self.d_ce_scale[imin] = (self.xu_scale - self.x_scale - self.d_c_scale)[imin]
            alpha_upper_max = self.divide_clip_zero(  self.xu_scale - self.x_scale - self.d_c_scale, self.d_ce_scale, cutoff=1., positive = True )
            
        alpha_eqp_max = min(alpha_lower_max, alpha_upper_max)
        
        alpha_eqp = alpha_eqp_max
        q_dir_c_c = self.hess_lgn_prod_c_scale.dot(self.d_c_scale)
        q_dir_c_ce = self.hess_lgn_prod_ce_scale.dot(self.d_c_scale)
        q_dir_ce_ce = self.hess_lgn_prod_ce_scale.dot(self.d_ce_scale)
        
        q = self.merit_linearization(self.d_c_scale + alpha_eqp * self.d_ce_scale) + .5 * ( q_dir_c_c + 2 * alpha_eqp * q_dir_c_ce + alpha_eqp * alpha_eqp * q_dir_ce_ce )
        
        # the very basic line search we would like to have is q(d_c + alpha_eqp * d_ce) < q(d_c)
        # perform armijo backtracking to useful descend
        q_c = self.merit_linearization(self.d_c_scale) + .5 * q_dir_c_c
        downhill = self.merit_linearization_grad_prod(self.d_c_scale, self.d_ce_scale) + q_dir_c_c
        
        if self.logfile:
            self.file.write('q_cc, cce, cece:   %e, %e, %e\n' % (q_dir_c_c, q_dir_c_ce, q_dir_ce_ce) )
            self.file.write('q, q_c, downhill:  %e, %e, %e\n' % (q, q_c, downhill) )
        
        while q > q_c + self.eta_eqp * alpha_eqp * downhill and alpha_eqp >= self.alphamin and q > q_c:
            alpha_eqp *= self.tau
            q = self.merit_linearization(self.d_c_scale + alpha_eqp * self.d_ce_scale) + .5 * ( q_dir_c_c + 2 * alpha_eqp * q_dir_c_ce + alpha_eqp * alpha_eqp * q_dir_ce_ce )
        
        # set alpha_eqp to 0 if it becomes very small (minimizer at Cauchy point)
        if alpha_eqp < self.alphamin:
            alpha_eqp = 0
            q = q_c
            
        if self.logfile:
            self.file.write('alpha_eqp, max, q: %e, %e, %e\n' % (alpha_eqp, alpha_eqp_max, q) )
 
        if self.verbosity > 0:
            print "%7.1e" % alpha_eqp if alpha_eqp == alpha_eqp_max else self.colour_string("%7.1e" % alpha_eqp, colour='red'),  
            sys.stdout.flush()
            
        if alpha_eqp < alpha_eqp_max:
                self.statistics['alpha_qp_nf'] += 1
            
        self.d_t_scale = self.d_c_scale + alpha_eqp * self.d_ce_scale
            
        self.pred = self.merit_scale - q
        self.statistics['direction_trial'] += time.clock() - self.lasttime; self.lasttime = time.clock()                          
        
        # evaluate at trial point
        self.statistics['nobj'] += 1
        self.statistics['ncons'] += 1
        self.x_t_scale = self.x_scale + self.d_t_scale
        self.x_t = self.scale_var_inv.dot(self.x_t_scale)
        [self.obj_t, self.grad_t, self.cons_t, self.jac_cons_t] = self.firstorder(self.x_t, self.userdata)
        if not np.isfinite(self.obj_t).all():
            self.terminate = -290
            return
        if self.obj_t <= - self.objrange:
            self.terminate = 30
            return
        if not np.isfinite(self.grad_t).all():
            self.terminate = -291
            return
        if not np.isfinite(self.cons_t).all():
            self.terminate = -292
            return
        if not np.isfinite(self.jac_cons_t.data).all():
            self.terminate = -293
            return
        self.statistics['eval'] += time.clock() - self.lasttime; self.lasttime = time.clock()
        self.obj_t_scale = self.scale_obj * self.obj_t
        self.grad_t_scale = self.scale_obj * self.scale_var_inv.dot(self.grad_t)
        self.cons_t_scale = self.scale_cons.dot(self.cons_t)
        self.jac_cons_t_scale = self.scale_cons.dot(self.jac_cons_t.dot(self.scale_var_inv))
        self.merit_eval(trial=True)
        self.ared = self.merit_scale - self.merit_t_scale
        
        self.statistics['direction_trial'] += time.clock() - self.lasttime; self.lasttime = time.clock()                          
        
        
    def estimate_multiplier(self):
        """
        Compute least square estimation of lagrange multipliers,
        modified multipliers used in EQP subproblem
        and modified gradient used in EQP subproblem
        """

        # estimation of least square multipliers
        self.multiplier_scale = np.zeros(self.m)
        self.multiplier_simple_scale = np.zeros(self.n)
        m = saddlepointsolve.solve_saddlepoint_get_t_zero_t(self.n, self.m_active_simple, self.grad_scale)
        self.multiplier_scale[self.active_cons_lower] = m[:self.active_cons_lower.shape[0]]
        self.multiplier_scale[self.inequalities] = self.multiplier_scale[self.inequalities].clip(min=0.)
        self.multiplier_scale[self.active_cons_upper] = m[self.active_cons_lower.shape[0]:self.m_active].clip(max=0.)
        
        self.multiplier_simple_scale[self.active_var_lower] = m[self.m_active:self.m_active+self.active_var_lower.shape[0]].clip(0)
        self.multiplier_simple_scale[self.active_var_upper] = - m[self.m_active+self.active_var_lower.shape[0]:].clip(0)
        if self.defacto_simple_eq.shape[0] > 0:
            defacto_eq_lower = [ii for ii in range(self.active_var_lower.shape[0]) if self.active_var_lower[ii] in self.defacto_simple_eq]
            defacto_eq_upper = [self.active_var_lower.shape[0] + ii for ii in range(self.active_var_upper.shape[0]) if self.active_var_upper[ii] in self.defacto_simple_eq]
            self.multiplier_simple_scale[np.intersect1d(self.active_var_lower, self.defacto_simple_eq)] = m[self.m_active:][defacto_eq_lower]
            self.multiplier_simple_scale[np.intersect1d(self.active_var_upper, self.defacto_simple_eq)] = - m[self.m_active:][defacto_eq_upper]
        
        # modified multipliers and modified gradient for EQP subproblem
        self.multiplier_eqp_scale[:] = self.multiplier_scale[:]
        self.grad_eqp_scale[:] = self.grad_scale[:]
        viol = self.jac_cons_scale.dot(self.d_n_scale) + self.cons_scale
        for ii in range(self.m):
            if not (ii in self.active_cons_lower or ii in self.active_cons_upper):
                # it should not be necessary to distinguish between equalities and inequalities
                ## equalities:
                #if self.cl[ii] == self.cu[ii]: 
                #    if abs(viol[ii] - self.cl_scale[ii]) > 0:
                #        self.multiplier_eqp_scale[ii] = - self.penalty * np.sign(viol[ii] - self.cl_scale[ii])
                #        self.grad_eqp_scale += self.penalty * np.sign(viol[ii]) * np.array(self.jac_cons_scale.getrow(ii).todense())[0,:]
                ## inequalities
                #else:
                # there is at most one violation
                # test violation of lower inequality c(x) - c_lo >= 0
                if viol[ii] - self.cl_scale[ii] < 0:
                    self.multiplier_eqp_scale[ii] = self.penalty
                    self.grad_eqp_scale -= self.penalty * np.array(self.jac_cons_scale.getrow(ii).todense())[0,:]
                # test violation of upper inequality c(x) - c_up <= 0
                if viol[ii] - self.cu_scale[ii] > 0:
                    self.multiplier_eqp_scale[ii] = - self.penalty
                    self.grad_eqp_scale += self.penalty * np.array(self.jac_cons_scale.getrow(ii).todense())[0,:]
                    
        if self.logfile:
            self.file.write('Multiplier:        '); np.savetxt(self.file, self.multiplier_scale, newline=" "); self.file.write('\n')
            self.file.write('Multiplier EQP:    '); np.savetxt(self.file, self.multiplier_eqp_scale, newline=" "); self.file.write('\n')
            self.file.write('grad EQP mod:      '); np.savetxt(self.file, self.grad_eqp_scale, newline=" "); self.file.write('\n')
      
        self.statistics['nhessv'] += 1
        self.statistics['multiplier'] += time.clock() - self.lasttime; self.lasttime = time.clock()              
        self.grad_eqp_scale += self.scale_var_inv.dot(self.hess_lgn_prod(self.x, self.scale_obj, -self.scale_cons.dot(self.multiplier_eqp_scale), self.scale_var_inv.dot(self.d_n_scale), self.userdata ))
        if not np.isfinite(self.grad_eqp_scale).all():
            self.terminate = -294
            return -1
        self.statistics['eval'] += time.clock() - self.lasttime; self.lasttime = time.clock()
        self.grad_eqp_scale -= self.multiplier_simple_scale
        self.statistics['multiplier'] += time.clock() - self.lasttime; self.lasttime = time.clock()
        
        if self.logfile:
            self.file.write('grad EQP total:    '); np.savetxt(self.file, self.grad_eqp_scale, newline=" "); self.file.write('\n')
        return 0
        
    def second_order_correction(self):
        """
        Second Order Correction
        """
        # compute d_soc such that A*d_soc + working constraints at trial point = 0
        rhs_cons = np.empty([self.m_active])
        rhs_cons[:self.active_cons_lower.shape[0]] = self.cons_t_scale[self.active_cons_lower] - self.cl_scale[self.active_cons_lower]
        rhs_cons[self.active_cons_lower.shape[0]:] = self.cons_t_scale[self.active_cons_upper] - self.cu_scale[self.active_cons_upper]
        
        rhs = - np.concatenate((rhs_cons,
                               self.x_scale[self.active_var_lower] + self.d_t_scale[self.active_var_lower] - self.xl_scale[self.active_var_lower],
                               self.xu_scale[self.active_var_upper] - self.d_t_scale[self.active_var_upper] - self.x_scale[self.active_var_upper]))
        d_soc = saddlepointsolve.solve_saddlepoint_get_s_zero_s(self.n, self.m_active_simple, rhs)
        
        if self.logfile:
            self.file.write('d_soc:             '); np.savetxt(self.file, d_soc, newline=" "); self.file.write('\n')
        
        # find steplength alpha_soc such that simple bounds are satisfied
        alpha_soc = self.divide_clip_zero(  self.xu_scale - self.x_scale - self.d_t_scale, d_soc, cutoff=1., positive = True )
        alpha_soc = min(alpha_soc, self.divide_clip_zero( -self.xl_scale + self.x_scale + self.d_t_scale, d_soc, cutoff=1., positive = False ))
        self.d_t_scale = self.d_t_scale + alpha_soc * d_soc
        self.statistics['direction_soc'] += time.clock() - self.lasttime; self.lasttime = time.clock()
        
        if self.logfile:
            self.file.write('d_t_soc:           '); np.savetxt(self.file, self.d_t_scale, newline=" "); self.file.write('\n')
        
        # evaluate at corrected trial point
        self.statistics['nobj'] += 1
        self.statistics['ncons'] += 1
        self.x_t_scale = self.x_scale + self.d_t_scale
        self.x_t = self.scale_var_inv.dot(self.x_t_scale)
        [self.obj_t, self.grad_t, self.cons_t, self.jac_cons_t] = self.firstorder(self.x_t, self.userdata)
        if not np.isfinite(self.obj_t).all():
            self.terminate = -290
            return -1
        if self.obj_t <= - self.objrange:
            self.terminate = 30
            return
        if not np.isfinite(self.grad_t).all():
            self.terminate = -291
            return -1
        if not np.isfinite(self.cons_t).all():
            self.terminate = -292
            return -1
        if not np.isfinite(self.jac_cons_t.data).all():
            self.terminate = -293
            return -1
        self.statistics['eval'] += time.clock() - self.lasttime; self.lasttime = time.clock()
        self.obj_t_scale = self.scale_obj * self.obj_t
        self.grad_t_scale = self.scale_obj * self.scale_var_inv.dot(self.grad_t)
        self.cons_t_scale = self.scale_cons.dot(self.cons_t)
        self.jac_cons_t_scale = self.scale_cons.dot(self.jac_cons_t.dot(self.scale_var_inv))
        self.merit_eval(trial=True)
        self.ared = self.merit_scale - self.merit_t_scale
        merit_linearized = self.merit_linearization(self.d_t_scale)
        self.statistics['direction_soc'] += time.clock() - self.lasttime; self.lasttime = time.clock()                          
        self.statistics['nhessv'] += 1
        d_t = self.scale_var_inv.dot(self.d_t_scale)
        q_dir_t_t = self.hess_lgn_prod(self.x, self.scale_obj, -self.scale_cons.dot(self.multiplier_scale), d_t, self.userdata).dot(d_t)
        if not np.isfinite(q_dir_t_t).all():
            self.terminate = -294
            return -1
        self.statistics['eval'] += time.clock() - self.lasttime; self.lasttime = time.clock()              
        
        q = merit_linearized + .5 * q_dir_t_t
        self.pred = self.merit_scale - q
        
        self.statistics['direction_soc'] += time.clock() - self.lasttime; self.lasttime = time.clock()
        
        return 0
           
    def update_tr(self):
        """
        Updates Trust Region Radii
        """
        norm_t = np.linalg.norm(self.d_t_scale)
        norm_t_inf = np.linalg.norm(self.d_t_scale, np.inf)
        norm_c_inf = np.linalg.norm(self.d_c_scale, np.inf)

        if self.tr_ratio >= .9:
            self.tr_eqp = max(self.tr_eqp, 7*norm_t)
        elif self.tr_ratio >= .3:
            self.tr_eqp = max(self.tr_eqp, 2*norm_t)
        elif self.tr_ratio < self.tr_acceptance:
            self.tr_eqp = min(.5*self.tr_eqp, .95*norm_t) # this is .5*min(tr_eqp, norm_t) in Byrd
                                                          # consider only .5*tr_eqp
            
        if self.tr_ratio >= self.tr_acceptance:
            max_expr = max( 1.2*norm_t_inf, 1.2*norm_c_inf, .1*self.tr_lp )
            if self.cauchy_full:
                self.tr_lp = min(max_expr, 7*self.tr_lp)
            else:
                self.tr_lp = min(max_expr, self.tr_lp)
        else:
            self.tr_lp = min(self.tr_lp, max( .5*norm_t_inf, .1*self.tr_lp ))
           
    def fixinfnorm(self, x, ord):
        """
        Computes ||x||_inf with proper handling of x = []
        hopefully this will be fixed in numpy soon
        """
        if x.shape[0] == 0:
            return 0
        else:
            return np.linalg.norm(x, np.inf)
            
    def termination_criterion(self):
        """
        Computes termination criterion as described in Knitro 9.0 Manual
        """
        
        # account for active simple bounds which we excluded from the lagrangian
        res = self.grad_scale - self.jac_cons_scale.transpose().dot(self.multiplier_scale) - self.multiplier_simple_scale
        residue_lower   = self.cons_scale - self.cl_scale
        residue_upper   = self.cu_scale - self.cons_scale
        violation = np.minimum(residue_lower, residue_upper)
        compl     = np.multiply(violation, self.multiplier_scale)
    
        if self.logfile:
            self.file.write('grad - J^T*lambda:   '); np.savetxt(self.file, res, newline=" "); self.file.write('\n')
            self.file.write('violation:           '); np.savetxt(self.file, violation, newline=" "); self.file.write('\n')
            self.file.write('violation .* lambda: '); np.savetxt(self.file, compl, newline=" "); self.file.write('\n')
        
        self.opterr  = max( self.fixinfnorm(res, np.inf), self.fixinfnorm(compl, np.inf) )
        self.feaserr = max( self.fixinfnorm( residue_lower.clip(max=0.), np.inf ), self.fixinfnorm( residue_upper.clip(max=0.), np.inf ) )
        if self.m > 0:
            self.optscale = max(1.0, self.fixinfnorm(self.grad_scale, np.inf))
            self.optscale = max(self.optscale, 1+np.linalg.norm(self.multiplier_scale))
        else:
            self.optscale = max(1.0, min( self.obj_scale, self.fixinfnorm(self.grad_scale, np.inf) ) )
        self.feasscale = max(self.feasscale, 1+np.linalg.norm(self.x_scale))
        
        if self.opterr < max(self.optscale * self.opttol, self.opttol_abs) and self.feaserr < max(self.feasscale * self.feastol, self.feastol_abs):
            self.terminate = 0
            
        if self.terminate == -1 and self.fixinfnorm(self.d_t_scale, np.inf) < self.chtol and self.stepaccepted: # and self.pred > 0: # i do not know if one should include this last condition
            if self.feaserr < max(self.feasscale * self.feastol, self.feastol_abs):
                self.terminate = 11
            else:
                self.terminate = 21
                    
    def loop(self):
        """
        Perform SLEQP iteration
        """
        self.statistics['iter'] += 1
        
        if self.verbosity > 0:
            if self.statistics['iter'] % 25 == 1:
                print " iter|   " \
                    + self.colour_string(u"ν") + "|  " \
                    + self.colour_string(u"LP Δ") + "|  " \
                    + self.colour_string(u"LP d") + "|  " \
                    + u" LP α" + "| " \
                    + u"CG " + "|  " \
                    + self.colour_string(u"QP Δ") + "|  " \
                    + self.colour_string(u"QP d") + "|  " \
                    + u" QP α" + "|   " \
                    + self.colour_string(u"τ") + "| " \
                    + u"?|  " \
                    + self.colour_string(u"   d") + "|  " \
                    + u"        obj" + "|  " \
                    + u"  pen" + "|  " \
                    + u" merit" + "|  " \
                    + self.colour_string(u"oε a") + "|  " \
                    + self.colour_string(u"oε r") + "|  " \
                    + self.colour_string(u"fε a") + "|  " \
                    + self.colour_string(u"fε r") + "|" 
            print "%5d" % self.statistics['iter'],
            sys.stdout.flush()
            
        if self.logfile:
            self.file.write('Iteration:         %s\n' % self.statistics['iter'])
            self.file.write('x:                 '); np.savetxt(self.file, self.x, newline=" "); self.file.write('\n')
            self.file.write('x_scale:           '); np.savetxt(self.file, self.x_scale, newline=" "); self.file.write('\n')
            
        # get active set at current point via lp solution, ensure feasibility
        # by eventual penalty update
        self.lasttime = time.clock()
       
        if self.solve_lp() < 0:
            if self.verbosity > 1 and self.logfile != False:
                self.file = open(self.logfile, 'a')
            return
        
        self.merit_eval()
        self.statistics['auxiliary'] += time.clock() - self.lasttime; self.lasttime = time.clock()              
        
        if self.setup_active_constraints() < 0:
            return
        
        self.direction_n()
        if self.estimate_multiplier() < 0:
            return
        
        if self.direction_cauchy() < 0:
            return
        
        if self.direction_ce() < 0:
            return
        self.direction_trial()
        
        if self.logfile:
            self.file.write('Pred. Reduction:   %e\n' % self.pred)
            self.file.write('Actual Reduction:  %e\n' % self.ared)

        self.statistics['auxiliary'] += time.clock() - self.lasttime; self.lasttime = time.clock()
            
        if self.pred == 0.:
            self.socattempted = True
            if self.second_order_correction() < 0:
                return
            self.predzero = True
            if self.logfile:
                self.file.write('SOC Pred. Red:     %e\n' % self.pred)
                self.file.write('SOC Actual Red:    %e\n' % self.ared)
            if self.pred <= 0.:
                self.prednotpositivesoc = True
            if self.pred == 0.:
                if self.ared >= self.pred:
                    self.tr_ratio = .5
                else:
                    self.tr_ratio = 1e-10
            else:
                self.tr_ratio = self.ared/np.abs(self.pred)

        else:
            self.tr_ratio = self.ared/self.pred
            self.socattempted = False
            if self.tr_ratio < self.tr_acceptance:
                self.socattempted = True
                if self.second_order_correction() < 0 :
                    return
                if self.logfile:
                    self.file.write('SOC Pred. Red:     %e\n' % self.pred)
                    self.file.write('SOC Actual Red:    %e\n' % self.ared)
                if self.pred <= 0.:
                    self.prednotpositivesoc = True
                if self.pred == 0.:
                    if self.ared >= self.pred:
                        self.tr_ratio = .5
                    else:
                        self.tr_ratio = 1e-10
                else:
                    self.tr_ratio = self.ared/np.abs(self.pred)
        if self.logfile:
            self.file.write('TR Ratio:          %e\n' % self.tr_ratio)
                
        if self.verbosity > 0:
            if self.tr_ratio < 0:
                print self.colour_string(self.neglog10(-self.tr_ratio, format="%4.1f"), colour='red'),
            else:
                print self.neglog10(self.tr_ratio, format="%4.1f"), 
            sys.stdout.flush()
        
        self.update_tr()
        
        if not np.isfinite(self.tr_ratio):
            self.terminate = -207
            return
        
        if self.tr_ratio >= self.tr_acceptance:
            self.stepaccepted = True
            if self.verbosity > 0:
                if self.socattempted:
                    print self.colour_string(" +", colour='red'), 
                else:
                    print " +",
                print self.neglog10(np.linalg.norm(self.d_t_scale)),
                
            self.statistics['nsuccess'] += 1
            self.x = self.x_t
            self.x_scale = self.x_t_scale
            self.obj = self.obj_t
            self.obj_scale = self.obj_t_scale
            self.grad = self.grad_t
            self.grad_scale = self.grad_t_scale
            self.cons = self.cons_t
            self.cons_scale = self.cons_t_scale
            self.jac_cons = self.jac_cons_t
            self.jac_cons_scale = self.jac_cons_t_scale
            mnorm = self.fixinfnorm(self.multiplier_scale, np.inf)
            if self.penalty > 1000*(mnorm+1):
                self.penaltytolarge += 1
            if self.penaltytolarge >= 5:
                self.penalty = 10 + mnorm
        else:
            self.stepaccepted = False
            if self.verbosity > 0:
                if self.socattempted:
                    print self.colour_string(" -", colour='red'), 
                else:
                    print " -",
                print self.neglog10(0.),
                sys.stdout.flush()
            self.statistics['nfail'] += 1
            self.statistics['auxiliary'] += time.clock() - self.lasttime; self.lasttime = time.clock()

        if self.verbosity > 0:
            print "%13.6e" % self.obj_scale,
            print "%7.1e" % (self.merit_scale - self.obj_scale),
            print "%7.1e" % self.merit_t_scale,  
            sys.stdout.flush()
            
        if self.logfile:
            self.file.write('Objective:         %e\n' % self.obj_scale)
            self.file.write('Penalty:           %e\n' % (self.merit_scale - self.obj_scale))
            self.file.write('Actual Reduction:  %e\n' % self.merit_scale)

        self.statistics['auxiliary'] += time.clock() - self.lasttime; self.lasttime = time.clock()
        self.termination_criterion()
 
        self.statistics['auxiliary'] += time.clock() - self.lasttime; self.lasttime = time.clock()
        
        if self.verbosity > 0:
            print self.neglog10(self.opterr),
            if self.opterr < max(self.optscale * self.opttol, self.opttol_abs):
                print self.colour_string( self.neglog10(self.opterr/self.optscale), colour='green' ),
            else:
                print self.neglog10(self.opterr/self.optscale),
            print self.neglog10(self.feaserr),
            if self.feaserr < max(self.feasscale * self.feastol, self.feastol_abs):
                print self.colour_string(self.neglog10(self.feaserr/self.feasscale), colour='green'),
            else:
                print self.neglog10(self.feaserr/self.feasscale),
            print 
            sys.stdout.flush()
            
        if self.logfile:
            self.file.write('Opterr  < :  %e, %e\n'   % (self.opterr,  max(self.optscale * self.opttol,   self.opttol_abs ) ) )
            self.file.write('Feaserr < :  %e, %e\n\n' % (self.feaserr, max(self.feasscale * self.feastol, self.feastol_abs) ) )
           
    def scale_problem(self, scale_var=None, scale_cons=None):
        self.scale_obj = 1.0
        self.scale_obj_inv = 1.0
        if scale_var == None:
            self.scale_var = scipy.sparse.eye(self.n)
            self.scale_var_inv = scipy.sparse.eye(self.n)
        else:
            self.scale_var = scale_var[0]
            self.scale_var_inv = scale_var[1]
        if scale_cons == None:
            self.scale_cons = scipy.sparse.eye(self.m)
            self.scale_cons_inv = scipy.sparse.eye(self.m)
        else:
            self.scale_cons = scale_cons[0]
            self.scale_cons_inv = scale_cons[1]            
        self.x_scale = self.scale_var.dot(self.x)
        self.xl_scale = self.scale_var.dot(self.xl)
        self.xu_scale = self.scale_var.dot(self.xu)
        self.cl_scale = self.scale_cons.dot(self.cl)
        self.cu_scale = self.scale_cons.dot(self.cu)
            
    def optimize(self, scale_var = None, scale_cons=None):
        """
        Call Optimization loop
        """
        if not self.stdoutset:
            if sys.stdout.isatty():
                default_encoding = sys.stdout.encoding
            else:
                default_encoding = locale.getpreferredencoding()
            sys.stdout = codecs.getwriter(default_encoding)(sys.stdout)
        
        if np.any(self.xl > self.xu):
            self.terminate = 23
        if np.any(self.cl > self.cu):
            self.terminate = 24
        self.defacto_simple_eq = np.where(self.xl == self.xu)[0]
        self.equalities = np.where(self.cl == self.cu)[0]
        self.m_eq = len(self.equalities)
        self.inequalities = np.setdiff1d(np.arange(self.m), self.equalities)
        self.m_ineq = self.m - self.m_eq
        if self.m > 0:
            self.slacks = scipy.sparse.hstack( [ scipy.sparse.identity(self.m), - scipy.sparse.identity(self.m) ] )
        self.lasttime = time.clock()
        self.zerotime = time.clock()
        self.statistics['nobj'] += 1
        self.statistics['ncons'] += 1
        try:
            [self.obj, self.grad, self.cons, self.jac_cons] = self.firstorder(self.x, self.userdata)
            if not np.isfinite(self.obj).all():
                self.terminate = -290
            if self.obj <= - self.objrange:
                self.terminate = 30
            if not np.isfinite(self.grad).all():
                self.terminate = -291
            if not np.isfinite(self.cons).all():
                self.terminate = -292
            if not np.isfinite(self.jac_cons.data).all():
                self.terminate = -293
        except Exception, e:
            print traceback.format_exc()
            print e
            self.statistics['exception'] = repr(e)
            self.terminate = -200
        self.statistics['eval'] += time.clock() - self.lasttime; self.lasttime = time.clock()
                
        if self.copyrightandstat:
            print "SLEQP Prototype"
            print u" (c) 2013-2016 Felix Lenders, IWR, Universität Heidelberg"
            print
            print "Using %s, GLTR and MA57" % self.lpsolver.upper()
            print
            print "#Variables:                 %d" % self.n
            print "#Constraints = #eq + #ineq: %d = %d + %d" % (self.m, self.m_eq, self.m_ineq)
            print
            
        self.scale_problem(scale_var=scale_var, scale_cons=scale_cons)
        if self.terminate >= -1:
            self.obj_scale = self.scale_obj * self.obj
            self.grad_scale = self.scale_obj * self.scale_var_inv.dot(self.grad)
            self.cons_scale = self.scale_cons.dot(self.cons)
            self.jac_cons_scale = self.scale_cons.dot(self.jac_cons.dot(self.scale_var_inv))
        
        if self.lpsolver == 'gurobi':
            self.lpsolvelib = __import__('pygurobi')
        elif self.lpsolver == 'soplex':
            self.lpsolvelib = __import__('lpsolvesoplex')
        
        if np.any( self.xl_scale - self.x_scale > 0. ) or np.any( self.x_scale - self.xu_scale > 0. ):
            self.terminate = -1
            if self.verbosity > 0:
                print "Starting Point violates simple bound. Find one that is simple feasible...",
            self.solve_lp(simple_infeasible = True)
            if self.verbosity > 0:
                if not( np.any( self.xl_scale - self.x_scale > 0. ) or np.any( self.x_scale - self.xu_scale > 0. ) ):
                    print "found"
                else:
                    print "failed"
                    self.terminate = -204
            try:
                [self.obj, self.grad, self.cons, self.jac_cons] = self.firstorder(self.x, self.userdata)
                if not np.isfinite(self.obj).all():
                    self.terminate = -290
                if self.obj <= - self.objrange:
                    self.terminate = 30
                if not np.isfinite(self.grad).all():
                    self.terminate = -291
                if not np.isfinite(self.cons).all():
                    self.terminate = -292
                if not np.isfinite(self.jac_cons.data).all():
                    self.terminate = -293
            except Exception, e:
                print traceback.format_exc()
                print e
                self.statistics['exception'] = repr(e)
                self.terminate = -200
                
            if self.terminate >= -1:
                self.obj_scale = self.scale_obj * self.obj
                self.grad_scale = self.scale_obj * self.scale_var_inv.dot(self.grad)
                self.cons_scale = self.scale_cons.dot(self.cons)
                self.jac_cons_scale = self.scale_cons.dot(self.jac_cons.dot(self.scale_var_inv))
        
        if self.terminate >= -1:
            self.feasscale_init = max( 1.0, self.fixinfnorm( (self.cl_scale - self.cons_scale).clip(min=0.), np.inf ),\
                self.fixinfnorm( (self.cons_scale - self.cu_scale).clip(min=0.), np.inf ),\
                self.fixinfnorm( (self.xl_scale - self.x_scale).clip(min=0.), np.inf ),\
                self.fixinfnorm( (self.x_scale - self.xu_scale).clip(min=0.), np.inf ) )
        
        if self.logfile != False:
            self.file = open(self.logfile, 'w')
                   
        #while not self.terminate and self.statistics['iter'] < self.maxiter and time.clock() - self.zerotime < self.walltime:
        while self.terminate == -1:
            if self.statistics['nsuccess'] >= self.maxiter or self.statistics['nsuccess'] + self.statistics['nfail'] >= 2 * self.maxiter:
                self.terminate = -2
                continue
            if time.clock() - self.zerotime >= self.walltime:
                self.terminate = -3
                continue
            if self.tr_lp  < .01*self.chtol:
                self.terminate = -201
                continue
            if self.tr_eqp < .01*self.chtol:
                self.terminate = -202
                continue
            try:
                self.loop()
            except Exception, e:
                print traceback.format_exc()
                print e
                self.statistics['exception'] = repr(e)
                self.terminate = -200
                continue
            if self.callback != None:
                self.callback(self.x)
        
        if self.statistics['iter'] > 0:
            try:
                self.statistics['obj'] = self.obj_scale[0]
            except:
                pass
            try:
                self.statistics['infeas'] = (self.merit_scale[0] - self.obj_scale[0])/self.penalty
            except:
                pass

        self.statistics['terminate'] = self.terminate
        self.statistics['terminationmessage'] = self.terminatedict[self.statistics['terminate']]
        if self.statistics['terminate'] == -200:
            self.statistics['terminationmessage'] += ': ' + self.statistics['exception']
        if self.gltrproblem30:
            self.statistics['terminationmessage'] += ' CG Warning occured: Possible inaccurate solution found on boundary.'
        if self.gltrproblem40:
            self.statistics['terminationmessage'] += ' CG Warning occured: <p,Hp> = 0.'
        if self.predzero:
            self.statistics['terminationmessage'] += ' Trust Region Warning occured: Predicted Reduction 0 at one point.'
        if self.prednotpositivesoc:
            self.statistics['terminationmessage'] += ' Trust Region Warning occured: Predicted Reduction negative in Second Order Correction at one point.'
        self.statistics['totaltime'] = self.statistics['auxiliary'] + self.statistics['lp_determination']  
        self.statistics['totaltime'] += self.statistics['saddlepoint_factorization'] + self.statistics['saddlepoint_setup']
        self.statistics['totaltime'] += self.statistics['multiplier']
        self.statistics['totaltime'] += self.statistics['direction_eqp'] + self.statistics['direction_n']
        self.statistics['totaltime'] += self.statistics['direction_cauchy'] + self.statistics['direction_trial'] + self.statistics['direction_soc']
        self.statistics['solver'] = self.statistics['totaltime']
        self.statistics['totaltime'] += self.statistics['eval']
        self.statistics['opterrabs'] = self.opterr
        self.statistics['feaserrabs'] = self.feaserr
        try:
            self.statistics['opterrrel'] =  self.opterr/self.optscale[0]
        except:
            self.statistics['opterrrel'] =  self.opterr/self.optscale
        self.statistics['feaserrrel'] = self.feaserr/self.feasscale
        self.statistics['n'] = self.n
        self.statistics['m'] = self.m
        self.statistics['meq'] = self.m_eq
        self.statistics['mineq'] = self.m_ineq

        
        #if self.compute_cond and self.m > 0:
        #    self.statistics['cond_jac_cons'] = self.Working_Set.read_kappa()
        
        if self.copyrightandstat:
            statstring = []
            statstring.append("")
            statstring.append("Termination Status, Objective:       %s (%s), %e" % (self.statistics['terminate'], self.statistics['terminationmessage'], self.obj_scale if self.obj_scale else 0.0) )
            statstring.append("Opt Err | Feas Err (abs, rel):       %e, %e | %e, %e" % (self.statistics['opterrabs'], self.statistics['opterrrel'], self.statistics['feaserrabs'], self.statistics['feaserrrel']) )
            statstring.append("#Variables:                          %d" % self.n)
            statstring.append("#Constraints = #eq + #ineq:          %d = %d + %d" % (self.m, self.m_eq, self.m_ineq))
            statstring.append("#Iterations = #acc + #rej:           %d = %d + %d" % (self.statistics['iter'], self.statistics['nsuccess'], self.statistics['nfail']))
            statstring.append("#Eval Obj&Grd, Con&Jac, Hess*Vector: %d, %d, %d" % (self.statistics['nobj'], self.statistics['ncons'], self.statistics['nhessv']) )
            statstring.append("Total CPU Time = Eval + Solver:      %e = %e + %e" % (self.statistics['totaltime'], self.statistics['eval'], self.statistics['solver']))
            if self.statistics['totaltime'] > 0:        
                statstring.append("%%Eval:                               %s" % format(self.statistics['eval']/self.statistics['totaltime'],"%")) # , %s, %s, %s" (
                statstring.append("%%LP, %%LA, %%EQP, %%Multiplier:         %s, %s, %s, %s" % (\
                   format(self.statistics['lp_determination']/self.statistics['totaltime'],"%"),\
                   format((self.statistics['saddlepoint_factorization'] + self.statistics['saddlepoint_setup'])/self.statistics['totaltime'], "%"),\
                   format(self.statistics['direction_eqp']/self.statistics['totaltime'],"%"),\
                   format(self.statistics['multiplier']/self.statistics['totaltime'],"%") ) )
                statstring.append("%%dC, %%dN, %%dT, %%dSOC, %%Aux:          %s, %s, %s, %s, %s" % (\
                   format(self.statistics['direction_cauchy']/self.statistics['totaltime'],"%"),\
                   format(self.statistics['direction_n']/self.statistics['totaltime'],"%"),\
                   format(self.statistics['direction_trial']/self.statistics['totaltime'],"%"),\
                   format(self.statistics['direction_soc']/self.statistics['totaltime'],"%"),\
                   format(self.statistics['auxiliary']/self.statistics['totaltime'],"%") ) )
            
            print "\n".join(statstring)
            if self.logfile:
                self.file.write("\n".join(statstring))

        if self.logfile:
           self.file.close()
